# -*- coding: utf-8 -*-
"""Parsing of VCF files from ``str``
"""

import ast
import functools
import math
import re
import warnings

from . import header
from . import record
from . import exceptions
from .exceptions import (
    CannotConvertValue,
    LeadingTrailingSpaceInKey,
    UnknownFilter,
    UnknownVCFVersion,
    SpaceInChromLine,
)


from .compat import OrderedDict

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


# expected "#CHROM" header prefix when there are samples
REQUIRE_SAMPLE_HEADER = ("#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO", "FORMAT")
# expected "#CHROM" header prefix when there are no samples
REQUIRE_NO_SAMPLE_HEADER = ("#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO")

#: Supported VCF versions, a warning will be issued otherwise
SUPPORTED_VCF_VERSIONS = ("VCFv4.0", "VCFv4.1", "VCFv4.2", "VCFv4.3")


class QuotedStringSplitter:
    """Helper class for splitting quoted strings

    Has support for interpreting quoting strings but also brackets.  Meant
    for splitting the VCF header line dicts
    """

    #: state constant for normal
    NORMAL = 0
    #: state constant for quoted
    QUOTED = 1
    #: state constant for delimiter
    ESCAPED = 2
    #: state constant for array
    ARRAY = 3
    #: state constant for delimiter
    DELIM = 4

    def __init__(self, delim=",", quote='"', brackets="[]"):
        #: string delimiter
        self.delim = delim
        #: quote character
        self.quote = quote
        #: two-character string with opening and closing brackets
        assert len(brackets) == 2
        self.brackets = brackets

    def run(self, s):
        """Split string ``s`` at delimiter, correctly interpreting quotes

        Further, interprets arrays wrapped in one level of ``[]``.  No
        recursive brackets are interpreted (as this would make the grammar
        non-regular and currently this complexity is not needed).  Currently,
        quoting inside of braces is not supported either.  This is just to
        support the example from VCF v4.3.
        """
        begins, ends = [0], []
        # transition table
        DISPATCH = {
            self.NORMAL: self._handle_normal,
            self.QUOTED: self._handle_quoted,
            self.ARRAY: self._handle_array,
            self.DELIM: self._handle_delim,
            self.ESCAPED: self._handle_escaped,
        }
        # run state automaton
        state = self.NORMAL
        for pos, c in enumerate(s):
            state = DISPATCH[state](c, pos, begins, ends)
        ends.append(len(s))
        assert len(begins) == len(ends)
        # Build resulting list
        return [s[start:end] for start, end in zip(begins, ends)]

    def _handle_normal(self, c, pos, begins, ends):  # pylint: disable=W0613
        if c == self.delim:
            ends.append(pos)
            return self.DELIM
        elif c == self.quote:
            return self.QUOTED
        elif c == self.brackets[0]:
            return self.ARRAY
        else:
            return self.NORMAL

    def _handle_quoted(self, c, pos, begins, ends):  # pylint: disable=W0613
        if c == "\\":
            return self.ESCAPED
        elif c == self.quote:
            return self.NORMAL
        else:
            return self.QUOTED

    def _handle_array(self, c, pos, begins, ends):  # pylint: disable=W0613
        if c == self.brackets[1]:
            return self.NORMAL
        else:
            return self.ARRAY

    def _handle_delim(self, c, pos, begins, ends):  # pylint: disable=W0613
        begins.append(pos)
        return self.NORMAL

    def _handle_escaped(self, c, pos, begins, ends):  # pylint: disable=W0613
        return self.QUOTED


def split_quoted_string(s, delim=",", quote='"', brackets="[]"):
    return QuotedStringSplitter(delim, quote, brackets).run(s)


def split_mapping(pair_str):
    """Split the ``str`` in ``pair_str`` at ``'='``

    Warn if key needs to be stripped
    """
    orig_key, value = pair_str.split("=", 1)
    key = orig_key.strip()
    if key != orig_key:
        warnings.warn(
            "Mapping key {} has leading or trailing space".format(repr(orig_key)),
            LeadingTrailingSpaceInKey,
        )
    return key, value


def parse_mapping(value):
    """Parse the given VCF header line mapping

    Such a mapping consists of "key=value" pairs, separated by commas and
    wrapped into angular brackets ("<...>").  Strings are usually quoted,
    for certain known keys, exceptions are made, depending on the tag key.
    this, however, only gets important when serializing.

    :raises: :py:class:`vcfpy.exceptions.InvalidHeaderException` if
        there was a problem parsing the file
    """
    if not value.startswith("<") or not value.endswith(">"):
        raise exceptions.InvalidHeaderException(
            "Header mapping value was not wrapped in angular brackets"
        )
    # split the comma-separated list into pairs, ignoring commas in quotes
    pairs = split_quoted_string(value[1:-1], delim=",", quote='"')
    # split these pairs into key/value pairs, converting flags to mappings
    # to True
    key_values = []
    for pair in pairs:
        if "=" in pair:
            key, value = split_mapping(pair)
            if value.startswith('"') and value.endswith('"'):
                value = ast.literal_eval(value)
            elif value.startswith("[") and value.endswith("]"):
                value = [v.strip() for v in value[1:-1].split(",")]
        else:
            key, value = pair, True
        key_values.append((key, value))
    # return completely parsed mapping as OrderedDict
    return OrderedDict(key_values)


class HeaderLineParserBase:
    """Parse into appropriate HeaderLine"""

    def parse_key_value(self, key, value):
        """Parse the key/value pair

        :param str key: the key to use in parsing
        :param str value: the value to parse
        :returns: :py:class:`vcfpy.header.HeaderLine` object
        """
        raise NotImplementedError("Must be overridden")


class StupidHeaderLineParser(HeaderLineParserBase):
    """Parse into HeaderLine (no particular structure)"""

    def parse_key_value(self, key, value):
        return header.HeaderLine(key, value)


class MappingHeaderLineParser(HeaderLineParserBase):
    """Parse into HeaderLine (no particular structure)"""

    def __init__(self, line_class):
        """Initialize the parser"""
        #: the class to use for the VCF header line
        self.line_class = line_class

    def parse_key_value(self, key, value):
        return self.line_class(key, value, parse_mapping(value))


def build_header_parsers():
    """Return mapping for parsers to use for each VCF header type

    Inject the WarningHelper into the parsers.
    """
    result = {
        "ALT": MappingHeaderLineParser(header.AltAlleleHeaderLine),
        "contig": MappingHeaderLineParser(header.ContigHeaderLine),
        "FILTER": MappingHeaderLineParser(header.FilterHeaderLine),
        "FORMAT": MappingHeaderLineParser(header.FormatHeaderLine),
        "INFO": MappingHeaderLineParser(header.InfoHeaderLine),
        "META": MappingHeaderLineParser(header.MetaHeaderLine),
        "PEDIGREE": MappingHeaderLineParser(header.PedigreeHeaderLine),
        "SAMPLE": MappingHeaderLineParser(header.SampleHeaderLine),
        "__default__": StupidHeaderLineParser(),  # fallback
    }
    return result


# Field value converters
_CONVERTERS = {
    "Integer": int,
    "Float": float,
    "Flag": lambda x: True,
    "Character": str,
    "String": str,
}


def convert_field_value(type_, value):
    """Convert atomic field value according to the type"""
    if value == ".":
        return None
    elif type_ in ("Character", "String"):
        if "%" in value:
            for k, v in record.UNESCAPE_MAPPING:
                value = value.replace(k, v)
        return value
    else:
        try:
            return _CONVERTERS[type_](value)
        except ValueError:
            warnings.warn(
                ("{} cannot be converted to {}, keeping as " "string.").format(value, type_),
                CannotConvertValue,
            )
            return value


def parse_field_value(field_info, value):
    """Parse ``value`` according to ``field_info``"""
    if field_info.id == "FT":
        return [x for x in value.split(";") if x != "."]
    elif isinstance(value, bool) or field_info.type == "Flag":
        return True
    elif field_info.number == 1:
        return convert_field_value(field_info.type, value)
    else:
        if value == ".":
            return []
        else:
            return [convert_field_value(field_info.type, x) for x in value.split(",")]


# Regular expression for break-end
BREAKEND_PATTERN = re.compile("[\\[\\]]")


def parse_breakend(alt_str):
    """Parse breakend and return tuple with results, parameters for BreakEnd
    constructor
    """
    arr = BREAKEND_PATTERN.split(alt_str)
    mate_chrom, mate_pos = arr[1].split(":", 1)
    mate_pos = int(mate_pos)
    if mate_chrom[0] == "<":
        mate_chrom = mate_chrom[1:-1]
        within_main_assembly = False
    else:
        within_main_assembly = True
    FWD_REV = {True: record.FORWARD, False: record.REVERSE}
    orientation = FWD_REV[alt_str[0] == "[" or alt_str[0] == "]"]
    mate_orientation = FWD_REV["[" in alt_str]
    if orientation == record.FORWARD:
        sequence = arr[2]
    else:
        sequence = arr[0]
    return (mate_chrom, mate_pos, orientation, mate_orientation, sequence, within_main_assembly)


def process_sub_grow(ref, alt_str):
    """Process substution where the string grows"""
    if len(alt_str) == 0:
        raise exceptions.InvalidRecordException("Invalid VCF, empty ALT")
    elif len(alt_str) == 1:
        if ref[0] == alt_str[0]:
            return record.Substitution(record.DEL, alt_str)
        else:
            return record.Substitution(record.INDEL, alt_str)
    else:
        return record.Substitution(record.INDEL, alt_str)


def process_sub_shrink(ref, alt_str):
    """Process substution where the string shrink"""
    if len(ref) == 0:
        raise exceptions.InvalidRecordException("Invalid VCF, empty REF")
    elif len(ref) == 1:
        if ref[0] == alt_str[0]:
            return record.Substitution(record.INS, alt_str)
        else:
            return record.Substitution(record.INDEL, alt_str)
    else:
        return record.Substitution(record.INDEL, alt_str)


def process_sub(ref, alt_str):
    """Process substitution"""
    if len(ref) == len(alt_str):
        if len(ref) == 1:
            return record.Substitution(record.SNV, alt_str)
        else:
            return record.Substitution(record.MNV, alt_str)
    elif len(ref) > len(alt_str):
        return process_sub_grow(ref, alt_str)
    else:  # len(ref) < len(alt_str):
        return process_sub_shrink(ref, alt_str)


def process_alt(header, ref, alt_str):  # pylint: disable=W0613
    """Process alternative value using Header in ``header``"""
    # By its nature, this function contains a large number of case distinctions
    if "]" in alt_str or "[" in alt_str:
        return record.BreakEnd(*parse_breakend(alt_str))
    elif alt_str[0] == "." and len(alt_str) > 0:
        return record.SingleBreakEnd(record.FORWARD, alt_str[1:])
    elif alt_str[-1] == "." and len(alt_str) > 0:
        return record.SingleBreakEnd(record.REVERSE, alt_str[:-1])
    elif alt_str[0] == "<" and alt_str[-1] == ">":
        inner = alt_str[1:-1]
        return record.SymbolicAllele(inner)
    else:  # substitution
        return process_sub(ref, alt_str)


class HeaderParser:
    """Helper class for parsing a VCF header"""

    def __init__(self):
        #: Sub parsers to use for parsing the header lines
        self.sub_parsers = build_header_parsers()

    def parse_line(self, line):
        """Parse VCF header ``line`` (trailing '\r\n' or '\n' is ignored)

        :param str line: ``str`` with line to parse
        :param dict sub_parsers: ``dict`` mapping header line types to
            appropriate parser objects
        :returns: appropriate :py:class:`HeaderLine` parsed from ``line``
        :raises: :py:class:`vcfpy.exceptions.InvalidHeaderException` if
            there was a problem parsing the file
        """
        if not line or not line.startswith("##"):
            raise exceptions.InvalidHeaderException(
                'Invalid VCF header line (must start with "##") {}'.format(line)
            )
        if "=" not in line:
            raise exceptions.InvalidHeaderException(
                'Invalid VCF header line (must contain "=") {}'.format(line)
            )
        line = line[len("##") :].rstrip()  # trim '^##' and trailing whitespace
        # split key/value pair at "="
        key, value = split_mapping(line)
        sub_parser = self.sub_parsers.get(key, self.sub_parsers["__default__"])
        return sub_parser.parse_key_value(key, value)


class RecordParser:
    """Helper class for parsing VCF records"""

    def __init__(self, header, samples, record_checks=None):
        #: Header with the meta information
        self.header = header
        #: SamplesInfos with sample information
        self.samples = samples
        #: The checks to perform, can contain 'INFO' and 'FORMAT'
        self.record_checks = tuple(record_checks or [])
        # Expected number of fields
        if self.samples.names:
            self.expected_fields = 9 + len(self.samples.names)
        else:
            self.expected_fields = 8
        # Cache of FieldInfo objects by FORMAT string
        self._format_cache = {}
        # Cache of FILTER entries, also applied to FORMAT/FT
        self._filter_ids = set(self.header.filter_ids())
        # Helper for checking INFO fields
        if "INFO" in self.record_checks:
            self._info_checker = InfoChecker(self.header)
        else:
            self._info_checker = NoopInfoChecker()
        # Helper for checking FORMAT fields
        if "FORMAT" in self.record_checks:
            self._format_checker = FormatChecker(self.header)
        else:
            self._format_checker = NoopFormatChecker()

    def parse_line(self, line_str):
        """Parse line from file (including trailing line break) and return
        resulting Record
        """
        line_str = line_str.rstrip()
        if not line_str:
            return None  # empty line, EOF
        arr = self._split_line(line_str)
        # CHROM
        chrom = arr[0]
        # POS
        pos = int(arr[1])
        # IDS
        if arr[2] == ".":
            ids = []
        else:
            ids = arr[2].split(";")
        # REF
        ref = arr[3]
        # ALT
        alts = []
        if arr[4] != ".":
            for alt in arr[4].split(","):
                alts.append(process_alt(self.header, ref, alt))
        # QUAL
        if arr[5] == ".":
            qual = None
        else:
            try:
                qual = int(arr[5])
            except ValueError:  # try as float
                qual = float(arr[5])
        # FILTER
        if arr[6] == ".":
            filt = []
        else:
            filt = arr[6].split(";")
        self._check_filters(filt, "FILTER")
        # INFO
        info = self._parse_info(arr[7], len(alts))
        if len(arr) == 9:
            raise exceptions.IncorrectVCFFormat("Expected 8 or 10+ columns, got 9!")
        elif len(arr) == 8:
            format_ = None
            calls = None
        else:
            # FORMAT
            format_ = arr[8].split(":")
            # sample/call columns
            calls = self._handle_calls(alts, format_, arr[8], arr)
        return record.Record(chrom, pos, ids, ref, alts, qual, filt, info, format_, calls)

    def _handle_calls(self, alts, format_, format_str, arr):
        """Handle FORMAT and calls columns, factored out of parse_line"""
        if format_str not in self._format_cache:
            self._format_cache[format_str] = list(map(self.header.get_format_field_info, format_))
        # per-sample calls
        calls = []
        for sample, raw_data in zip(self.samples.names, arr[9:]):
            if self.samples.is_parsed(sample):
                data = self._parse_calls_data(format_, self._format_cache[format_str], raw_data)
                call = record.Call(sample, data)
                self._format_checker.run(call, len(alts))
                self._check_filters(call.data.get("FT"), "FORMAT/FT", call.sample)
                calls.append(call)
            else:
                calls.append(record.UnparsedCall(sample, raw_data))
        return calls

    def _check_filters(self, filt, source, sample=None):
        if not filt:
            return
        for f in filt:
            self._check_filter(f, source, sample)

    def _check_filter(self, f, source, sample):
        if f == "PASS":
            pass  # the PASS filter is implicitely defined
        elif f not in self._filter_ids:
            if source == "FILTER":
                warnings.warn(
                    ("Filter not found in header: {}; problem in FILTER " "column").format(f),
                    UnknownFilter,
                )
            else:
                assert source == "FORMAT/FT" and sample
                warnings.warn(
                    (
                        "Filter not found in header: {}; problem in "
                        "FORMAT/FT column of sample {}"
                    ).format(f, sample),
                    UnknownFilter,
                )

    def _split_line(self, line_str):
        """Split line and check number of columns"""
        arr = line_str.rstrip().split("\t")
        if len(arr) != self.expected_fields:
            raise exceptions.InvalidRecordException(
                (
                    "The line contains an invalid number of fields. Was "
                    "{} but expected {}\n{}".format(len(arr), 9 + len(self.samples.names), line_str)
                )
            )
        return arr

    def _parse_info(self, info_str, num_alts):
        """Parse INFO column from string"""
        result = OrderedDict()
        if info_str == ".":
            return result
        # The standard is very nice to parsers, we can simply split at
        # semicolon characters, although I (Manuel) don't know how strict
        # programs follow this
        for entry in info_str.split(";"):
            if "=" not in entry:  # flag
                key = entry
                result[key] = parse_field_value(self.header.get_info_field_info(key), True)
            else:
                key, value = split_mapping(entry)
                result[key] = parse_field_value(self.header.get_info_field_info(key), value)
            self._info_checker.run(key, result[key], num_alts)
        return result

    @classmethod
    def _parse_calls_data(klass, format_, infos, gt_str):
        """Parse genotype call information from arrays using format array

        :param list format: List of strings with format names
        :param gt_str arr: string with genotype information values
        """
        data = OrderedDict()
        # The standard is very nice to parsers, we can simply split at
        # colon characters, although I (Manuel) don't know how strict
        # programs follow this
        for key, info, value in zip(format_, infos, gt_str.split(":")):
            data[key] = parse_field_value(info, value)
        return data


class HeaderChecker:
    """Helper class for checking a VCF header"""

    def run(self, header):
        """Check the header

        Warnings will be printed using ``warnings`` while errors will raise
        an exception.

        :raises: ``vcfpy.exceptions.InvalidHeaderException`` in the case of
            severe errors reading the header
        """
        self._check_header_lines(header.lines)

    def _check_header_lines(self, header_lines):
        """Check header lines, in particular for starting file "##fileformat" """
        if not header_lines:
            raise exceptions.InvalidHeaderException(
                "The VCF file did not contain any header lines!"
            )
        first = header_lines[0]
        if first.key != "fileformat":
            raise exceptions.InvalidHeaderException("The VCF file did not start with ##fileformat")
        if first.value not in SUPPORTED_VCF_VERSIONS:
            warnings.warn("Unknown VCF version {}".format(first.value), UnknownVCFVersion)


@functools.lru_cache(maxsize=32)
def binomial(n, k):
    try:
        res = math.factorial(n) // math.factorial(k) // math.factorial(n - k)
    except ValueError:
        res = 0
    return res


class NoopInfoChecker:
    """Helper class that performs no checks"""

    def __init__(self):
        pass

    def run(self, key, value, num_alts):
        pass


class InfoChecker:
    """Helper class for checking an INFO field"""

    def __init__(self, header):
        #: VCFHeader to use for checking
        self.header = header

    def run(self, key, value, num_alts):
        """Check value in INFO[key] of record

        Currently, only checks for consistent counts are implemented

        :param str key: key of INFO entry to check
        :param value: value to check
        :param int alts: list of alternative alleles, for length
        """
        field_info = self.header.get_info_field_info(key)
        if not isinstance(value, list):
            return
        TABLE = {
            ".": len(value),
            "A": num_alts,
            "R": num_alts + 1,
            "G": binomial(num_alts + 1, 2),  # diploid only at the moment
        }
        expected = TABLE.get(field_info.number, field_info.number)
        if len(value) != expected:
            tpl = "Number of elements for INFO field {} is {} instead of {}"
            warnings.warn(
                tpl.format(key, len(value), field_info.number), exceptions.IncorrectListLength
            )


class NoopFormatChecker:
    """Helper class that performs no checks"""

    def __init__(self):
        pass

    def run(self, call, num_alts):
        pass


class FormatChecker:
    """Helper class for checking a FORMAT field"""

    def __init__(self, header):
        #: VCFHeader to use for checking
        self.header = header

    def run(self, call, num_alts):
        """Check ``FORMAT`` of a record.Call

        Currently, only checks for consistent counts are implemented
        """
        for key, value in call.data.items():
            self._check_count(call, key, value, num_alts)

    def _check_count(self, call, key, value, num_alts):
        field_info = self.header.get_format_field_info(key)
        if isinstance(value, list):
            return
        num_alleles = len(call.gt_alleles or [])
        TABLE = {
            ".": len(value),
            "A": num_alts,
            "R": num_alts + 1,
            "G": binomial(num_alts + num_alleles, num_alleles),
        }
        expected = TABLE.get(field_info.number, field_info.number)
        if len(value) != expected:
            tpl = (
                "Number of elements for FORMAT field {} is {} instead "
                "of {} (number specifier {})"
            )
            warnings.warn(
                tpl.format(key, len(value), expected, field_info.number),
                exceptions.IncorrectListLength,
            )


class Parser:
    """Class for line-wise parsing of VCF files

    In most cases, you want to use :py:class:`vcfpy.reader.Reader` instead.

    :param stream: ``file``-like object to read from
    :param str path: path the VCF is parsed from, for display purposes
        only, optional
    """

    def __init__(self, stream, path=None, record_checks=None):
        self.stream = stream
        self.path = path
        #: checks to perform, can contain 'INFO' and 'FORMAT'
        self.record_checks = tuple(record_checks or [])
        #: header, once it has been read
        self.header = None
        # the currently read line
        self._line = stream.readline()  # trailing '\n'
        #: :py:class:`vcfpy.header.SamplesInfos` with sample information;
        #: set on parsing the header
        self.samples = None
        # helper for parsing the records
        self._record_parser = None
        # helper for checking the header
        self._header_checker = HeaderChecker()

    def _read_next_line(self):
        """Read next line store in self._line and return old one"""
        prev_line = self._line
        self._line = self.stream.readline()
        return prev_line

    def parse_header(self, parsed_samples=None):
        """Read and parse :py:class:`vcfpy.header.Header` from file, set
        into ``self.header`` and return it

        :param list parsed_samples: ``list`` of ``str`` for subsetting the
            samples to parse
        :returns: ``vcfpy.header.Header``
        :raises: ``vcfpy.exceptions.InvalidHeaderException`` in the case of
            problems reading the header
        """
        # parse header lines
        sub_parser = HeaderParser()
        header_lines = []
        while self._line and self._line.startswith("##"):
            header_lines.append(sub_parser.parse_line(self._line))
            self._read_next_line()
        # parse sample info line
        self.samples = self._handle_sample_line(parsed_samples)
        # construct Header object
        self.header = header.Header(header_lines, self.samples)
        # check header for consistency
        self._header_checker.run(self.header)
        # construct record parser
        self._record_parser = RecordParser(self.header, self.samples, self.record_checks)
        # read next line, must not be header
        self._read_next_line()
        if self._line and self._line.startswith("#"):
            raise exceptions.IncorrectVCFFormat(
                'Expecting non-header line or EOF after "#CHROM" line'
            )
        return self.header

    def _handle_sample_line(self, parsed_samples=None):
        """ "Check and interpret the "##CHROM" line and return samples"""
        if not self._line or not self._line.startswith("#CHROM"):
            raise exceptions.IncorrectVCFFormat('Missing line starting with "#CHROM"')
        # check for space before INFO
        line = self._line.rstrip()
        pos = line.find("FORMAT") if ("FORMAT" in line) else line.find("INFO")
        if pos == -1:
            raise exceptions.IncorrectVCFFormat('Ill-formatted line starting with "#CHROM"')
        if " " in line[:pos]:
            warnings.warn(
                (
                    "Found space in #CHROM line, splitting at whitespace "
                    "instead of tab; this VCF file is ill-formatted"
                ),
                SpaceInChromLine,
            )
            arr = self._line.rstrip().split()
        else:
            arr = self._line.rstrip().split("\t")

        self._check_samples_line(arr)
        return header.SamplesInfos(arr[len(REQUIRE_SAMPLE_HEADER) :], parsed_samples)

    @classmethod
    def _check_samples_line(klass, arr):
        """Peform additional check on samples line"""
        if len(arr) <= len(REQUIRE_NO_SAMPLE_HEADER):
            if tuple(arr) != REQUIRE_NO_SAMPLE_HEADER:
                raise exceptions.IncorrectVCFFormat(
                    "Sample header line indicates no sample but does not "
                    "equal required prefix {}".format("\t".join(REQUIRE_NO_SAMPLE_HEADER))
                )
        elif tuple(arr[: len(REQUIRE_SAMPLE_HEADER)]) != REQUIRE_SAMPLE_HEADER:
            raise exceptions.IncorrectVCFFormat(
                'Sample header line (starting with "#CHROM") does not '
                "start with required prefix {}".format("\t".join(REQUIRE_SAMPLE_HEADER))
            )

    def parse_line(self, line):
        """Pare the given line without reading another one from the stream"""
        return self._record_parser.parse_line(line)

    def parse_next_record(self):
        """Read, parse and return next :py:class:`vcfpy.record.Record`

        :returns: next VCF record or ``None`` if at end
        :raises: ``vcfpy.exceptions.InvalidRecordException`` in the case of
            problems reading the record
        """
        return self.parse_line(self._read_next_line())

    def print_warn_summary(self):
        """If there were any warnings, print summary with warnings"""
        # TODO: remove?



# -*- coding: utf-8 -*-
"""Test parsing of full VCF header
"""

import io
import sys

import pytest

from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"

MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


@pytest.fixture(scope="function")
def medium_header():
    return io.StringIO(MEDIUM_HEADER)


def test_parse_header(medium_header):
    p = parser.Parser(stream=medium_header, path="<builtin>")
    header = p.parse_header()
    assert header.lines
    assert len(header.lines) == 18
    EXPECTED = "HeaderLine('fileformat', 'VCFv4.3')"
    assert str(header.lines[0]) == EXPECTED
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,"
            "Description=\"Haplotype Quality\">', OrderedDict([('ID', 'HQ'), "
            "('Number', 2), ('Type', 'Integer'), ('Description', "
            "'Haplotype Quality')]))"
        )
    else:
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=HQ,Number=2,Type=Integer,Description=\"Haplotype Quality\">', "
            "{'ID': 'HQ', 'Number': 2, 'Type': 'Integer', 'Description': 'Haplotype Quality'})"
        )
    assert str(header.lines[-1]) == EXPECTED
    assert header.samples.names == ["NA00001", "NA00002", "NA00003"]



# -*- coding: utf-8 -*-
"""Writing of VCF headers
"""

import io
import textwrap

import pytest

from vcfpy import header, parser, writer

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"

MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##ALT=<ID=R,Description="IUPAC code R = A/G">
##META=<ID=Assay,Type=String,Number=.,Values=[WholeGenome, Exome]>
##PEDIGREE=<ID=TumourSample,Original=GermlineID>
##SAMPLE=<ID=Sample1,Assay=WholeGenome,Ethnicity=AFR,Disease=None,Description="Patient germline genome from unaffected",DOI=url>
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


@pytest.fixture(scope="function")
def header_samples():
    p = parser.Parser(stream=io.StringIO(MEDIUM_HEADER), path="<builtin>")
    p.parse_header()
    return (p.header, p.samples)


def test_write_header(header_samples, tmpdir_factory):
    path = tmpdir_factory.mktemp("write_header").join("out.vcf")
    header, _ = header_samples
    w = writer.Writer.from_path(path, header)
    w.close()
    RESULT = path.read()
    EXPECTED = MEDIUM_HEADER
    assert RESULT == EXPECTED


def test_write_header_no_samples(tmpdir_factory):
    # Create header to write out from scratch
    hdr = header.Header(
        lines=[header.HeaderLine("fileformat", "VCFv4.0")], samples=header.SamplesInfos([])
    )
    # Write out header
    path = tmpdir_factory.mktemp("write_header").join("out.vcf")
    w = writer.Writer.from_path(path, hdr)
    w.close()
    # Compare result
    RESULT = path.read()
    EXPECTED = textwrap.dedent(
        r"""
    ##fileformat=VCFv4.0
    #CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO
    """
    ).lstrip()
    assert RESULT == EXPECTED



# -*- coding: utf-8 -*-
"""Test escaping on writing VCF records
"""

from vcfpy import writer
from vcfpy import header

# vcfpy.writer.format_atomic() ------------------------------------------------


def test_format_atomic_with_escape_info():
    EXPECTED = "%3B%3D%25%2C%0D%0A%09"
    RESULT = writer.format_atomic(";=%,\r\n\t", "INFO")
    assert EXPECTED == RESULT


def test_format_atomic_with_escape_format():
    EXPECTED = "%3A%3D%25%2C%0D%0A%09"
    RESULT = writer.format_atomic(":=%,\r\n\t", "FORMAT")
    assert EXPECTED == RESULT


def test_format_atomic_without_escape_info():
    EXPECTED = "This is a legal string:"
    RESULT = writer.format_atomic("This is a legal string:", "INFO")
    assert EXPECTED == RESULT


def test_format_atomic_without_escape_format():
    EXPECTED = "This is a legal string;"
    RESULT = writer.format_atomic("This is a legal string;", "FORMAT")
    assert EXPECTED == RESULT


# vcfpy.writer.format_value() -------------------------------------------------


def test_format_value_with_escape():
    EXPECTED = "%3A%3B%3D%25%2C%0D%0A%09,%25"
    RESULT = writer.format_value(header.FieldInfo("String", 2), (":;=%,\r\n\t", "%"), "INFO")
    assert EXPECTED == RESULT


def test_format_value_without_escape():
    EXPECTED = "This is a legal string,me too"
    RESULT = writer.format_value(
        header.FieldInfo("String", 2), ("This is a legal string", "me too"), "INFO"
    )
    assert EXPECTED == RESULT



# -*- coding: utf-8 -*-
"""Reading of VCF files "from the wild"

The aim here is to test whether they can be parsed at all or make the parser
crash.
"""

import os

import pytest

from vcfpy import reader

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


FILENAMES = [
    "1kg.sites.vcf",
    "1kg.vcf.gz",
    "bcftools.vcf",
    "contig_idonly.vcf",
    "example-4.0.vcf",
    "example-4.1-info-multiple-values.vcf",
    "example-4.1-ploidy.vcf",
    "example-4.1-sv.vcf",
    "example-4.1.vcf",
    "example-4.2.vcf",
    "freebayes.vcf",
    "gatk_26_meta.vcf",
    "gatk.vcf",
    "gonl.chr20.release4.gtc.vcf",
    "info-type-character.vcf",
    "issue-140-file1.vcf",
    "issue-140-file2.vcf",
    "issue-140-file3.vcf",
    "issue-16.vcf",
    "issue-201.vcf.gz",
    "issue-214.vcf",
    "issue_49.vcf",
    "metadata-whitespace.vcf",
    "mixed-filtering.vcf",
    "null_genotype_mono.vcf",
    "parse-meta-line.vcf",
    "samples-space.vcf",
    "samtools.vcf",
    "strelka.vcf",
    "string_as_flag.vcf",
    "tb.vcf.gz",
    "uncalled_genotypes.vcf",
    "walk_left.vcf",
    "walk_refcall.vcf",
]


@pytest.mark.parametrize("filename", FILENAMES)
def test_read_vcfs_from_the_wild(filename):
    path = os.path.join(os.path.dirname(__file__), "vcfs_from_the_wild", filename)
    r = reader.Reader.from_path(path)



# -*- coding: utf-8 -*-
"""Tests for reserved field handling in header
"""

from vcfpy import header

# INFO for (small) variants ---------------------------------------------------


def test_info_aa():
    assert header.RESERVED_INFO["AA"]
    assert header.RESERVED_INFO["AA"].type == "String"
    assert header.RESERVED_INFO["AA"].number == 1
    assert header.RESERVED_INFO["AA"].description


def test_info_ac():
    assert header.RESERVED_INFO["AC"]
    assert header.RESERVED_INFO["AC"].type == "Integer"
    assert header.RESERVED_INFO["AC"].number == "A"
    assert header.RESERVED_INFO["AC"].description


def test_info_ad():
    assert header.RESERVED_INFO["AD"]
    assert header.RESERVED_INFO["AD"].type == "Integer"
    assert header.RESERVED_INFO["AD"].number == "R"
    assert header.RESERVED_INFO["AD"].description


def test_info_adf():
    assert header.RESERVED_INFO["ADF"]
    assert header.RESERVED_INFO["ADF"].type == "Integer"
    assert header.RESERVED_INFO["ADF"].number == "R"
    assert header.RESERVED_INFO["ADF"].description


def test_info_adr():
    assert header.RESERVED_INFO["ADR"]
    assert header.RESERVED_INFO["ADR"].type == "Integer"
    assert header.RESERVED_INFO["ADR"].number == "R"
    assert header.RESERVED_INFO["ADR"].description


def test_info_an():
    assert header.RESERVED_INFO["AN"]
    assert header.RESERVED_INFO["AN"].type == "Integer"
    assert header.RESERVED_INFO["AN"].number == 1
    assert header.RESERVED_INFO["AN"].description


def test_info_bq():
    assert header.RESERVED_INFO["BQ"]
    assert header.RESERVED_INFO["BQ"].type == "Float"
    assert header.RESERVED_INFO["BQ"].number == 1
    assert header.RESERVED_INFO["BQ"].description


def test_info_cigar():
    assert header.RESERVED_INFO["CIGAR"]
    assert header.RESERVED_INFO["CIGAR"].type == "String"
    assert header.RESERVED_INFO["CIGAR"].number == "A"
    assert header.RESERVED_INFO["CIGAR"].description


def test_info_DP():
    assert header.RESERVED_INFO["DB"]
    assert header.RESERVED_INFO["DB"].type == "Flag"
    assert header.RESERVED_INFO["DB"].number == 0
    assert header.RESERVED_INFO["DB"].description


def test_info_dp():
    assert header.RESERVED_INFO["DP"]
    assert header.RESERVED_INFO["DP"].type == "Integer"
    assert header.RESERVED_INFO["DP"].number == 1
    assert header.RESERVED_INFO["DP"].description


def test_info_h2():
    assert header.RESERVED_INFO["H2"]
    assert header.RESERVED_INFO["H2"].type == "Flag"
    assert header.RESERVED_INFO["H2"].number == 0
    assert header.RESERVED_INFO["H2"].description


def test_info_h3():
    assert header.RESERVED_INFO["H3"]
    assert header.RESERVED_INFO["H3"].type == "Flag"
    assert header.RESERVED_INFO["H3"].number == 0
    assert header.RESERVED_INFO["H3"].description


def test_info_mq():
    assert header.RESERVED_INFO["MQ"]
    assert header.RESERVED_INFO["MQ"].type == "Integer"
    assert header.RESERVED_INFO["MQ"].number == 1
    assert header.RESERVED_INFO["MQ"].description


def test_info_mq0():
    assert header.RESERVED_INFO["MQ0"]
    assert header.RESERVED_INFO["MQ0"].type == "Integer"
    assert header.RESERVED_INFO["MQ0"].number == 1
    assert header.RESERVED_INFO["MQ0"].description


def test_info_ns():
    assert header.RESERVED_INFO["NS"]
    assert header.RESERVED_INFO["NS"].type == "Integer"
    assert header.RESERVED_INFO["NS"].number == 1
    assert header.RESERVED_INFO["NS"].description


def test_info_sb():
    assert header.RESERVED_INFO["SB"]
    assert header.RESERVED_INFO["SB"].type == "Integer"
    assert header.RESERVED_INFO["SB"].number == 4
    assert header.RESERVED_INFO["SB"].description


def test_info_somatic():
    assert header.RESERVED_INFO["SOMATIC"]
    assert header.RESERVED_INFO["SOMATIC"].type == "Flag"
    assert header.RESERVED_INFO["SOMATIC"].number == 0
    assert header.RESERVED_INFO["SOMATIC"].description


def test_info_validated():
    assert header.RESERVED_INFO["VALIDATED"]
    assert header.RESERVED_INFO["VALIDATED"].type == "Flag"
    assert header.RESERVED_INFO["VALIDATED"].number == 0
    assert header.RESERVED_INFO["VALIDATED"].description


def test_info_1000g():
    assert header.RESERVED_INFO["1000G"]
    assert header.RESERVED_INFO["1000G"].type == "Flag"
    assert header.RESERVED_INFO["1000G"].number == 0
    assert header.RESERVED_INFO["1000G"].description


# INFO for SVs ----------------------------------------------------------------


def test_info_imprecise():
    assert header.RESERVED_INFO["IMPRECISE"]
    assert header.RESERVED_INFO["IMPRECISE"].type == "Flag"
    assert header.RESERVED_INFO["IMPRECISE"].number == 0
    assert header.RESERVED_INFO["IMPRECISE"].description


def test_info_novel():
    assert header.RESERVED_INFO["NOVEL"]
    assert header.RESERVED_INFO["NOVEL"].type == "Flag"
    assert header.RESERVED_INFO["NOVEL"].number == 0
    assert header.RESERVED_INFO["NOVEL"].description


def test_info_end():
    assert header.RESERVED_INFO["END"]
    assert header.RESERVED_INFO["END"].type == "Integer"
    assert header.RESERVED_INFO["END"].number == 1
    assert header.RESERVED_INFO["END"].description


def test_info_svtype():
    assert header.RESERVED_INFO["SVTYPE"]
    assert header.RESERVED_INFO["SVTYPE"].type == "String"
    assert header.RESERVED_INFO["SVTYPE"].number == 1
    assert header.RESERVED_INFO["SVTYPE"].description


def test_info_svlen():
    assert header.RESERVED_INFO["SVLEN"]
    assert header.RESERVED_INFO["SVLEN"].type == "Integer"
    assert header.RESERVED_INFO["SVLEN"].number == 1
    assert header.RESERVED_INFO["SVLEN"].description


def test_info_cipos():
    assert header.RESERVED_INFO["CIPOS"]
    assert header.RESERVED_INFO["CIPOS"].type == "Integer"
    assert header.RESERVED_INFO["CIPOS"].number == 2
    assert header.RESERVED_INFO["CIPOS"].description


def test_info_ciend():
    assert header.RESERVED_INFO["CIEND"]
    assert header.RESERVED_INFO["CIEND"].type == "Integer"
    assert header.RESERVED_INFO["CIEND"].number == 2
    assert header.RESERVED_INFO["CIEND"].description


def test_info_homlen():
    assert header.RESERVED_INFO["HOMLEN"]
    assert header.RESERVED_INFO["HOMLEN"].type == "Integer"
    assert header.RESERVED_INFO["HOMLEN"].number == "."
    assert header.RESERVED_INFO["HOMLEN"].description


def test_info_homseq():
    assert header.RESERVED_INFO["HOMSEQ"]
    assert header.RESERVED_INFO["HOMSEQ"].type == "String"
    assert header.RESERVED_INFO["HOMSEQ"].number == "."
    assert header.RESERVED_INFO["HOMSEQ"].description


def test_info_bkptid():
    assert header.RESERVED_INFO["BKPTID"]
    assert header.RESERVED_INFO["BKPTID"].type == "String"
    assert header.RESERVED_INFO["BKPTID"].number == "."
    assert header.RESERVED_INFO["BKPTID"].description


def test_info_meinfo():
    assert header.RESERVED_INFO["MEINFO"]
    assert header.RESERVED_INFO["MEINFO"].type == "String"
    assert header.RESERVED_INFO["MEINFO"].number == 4
    assert header.RESERVED_INFO["MEINFO"].description


def test_info_metrans():
    assert header.RESERVED_INFO["METRANS"]
    assert header.RESERVED_INFO["METRANS"].type == "String"
    assert header.RESERVED_INFO["METRANS"].number == 4
    assert header.RESERVED_INFO["METRANS"].description


def test_info_dgvid():
    assert header.RESERVED_INFO["DGVID"]
    assert header.RESERVED_INFO["DGVID"].type == "String"
    assert header.RESERVED_INFO["DGVID"].number == 1
    assert header.RESERVED_INFO["DGVID"].description


def test_info_dbvarid():
    assert header.RESERVED_INFO["DBVARID"]
    assert header.RESERVED_INFO["DBVARID"].type == "String"
    assert header.RESERVED_INFO["DBVARID"].number == 1
    assert header.RESERVED_INFO["DBVARID"].description


def test_info_dbripid():
    assert header.RESERVED_INFO["DBRIPID"]
    assert header.RESERVED_INFO["DBRIPID"].type == "String"
    assert header.RESERVED_INFO["DBRIPID"].number == 1
    assert header.RESERVED_INFO["DBRIPID"].description


def test_info_mateid():
    assert header.RESERVED_INFO["MATEID"]
    assert header.RESERVED_INFO["MATEID"].type == "String"
    assert header.RESERVED_INFO["MATEID"].number == "."
    assert header.RESERVED_INFO["MATEID"].description


def test_info_parid():
    assert header.RESERVED_INFO["PARID"]
    assert header.RESERVED_INFO["PARID"].type == "String"
    assert header.RESERVED_INFO["PARID"].number == 1
    assert header.RESERVED_INFO["PARID"].description


def test_info_event():
    assert header.RESERVED_INFO["EVENT"]
    assert header.RESERVED_INFO["EVENT"].type == "String"
    assert header.RESERVED_INFO["EVENT"].number == 1
    assert header.RESERVED_INFO["EVENT"].description


def test_info_cilen():
    assert header.RESERVED_INFO["CILEN"]
    assert header.RESERVED_INFO["CILEN"].type == "Integer"
    assert header.RESERVED_INFO["CILEN"].number == 2
    assert header.RESERVED_INFO["CILEN"].description


def test_info_dpadj():
    assert header.RESERVED_INFO["DPADJ"]
    assert header.RESERVED_INFO["DPADJ"].type == "Integer"
    assert header.RESERVED_INFO["DPADJ"].number == "."
    assert header.RESERVED_INFO["DPADJ"].description


def test_info_cn():
    assert header.RESERVED_INFO["CN"]
    assert header.RESERVED_INFO["CN"].type == "Integer"
    assert header.RESERVED_INFO["CN"].number == 1
    assert header.RESERVED_INFO["CN"].description


def test_info_cnadj():
    assert header.RESERVED_INFO["CNADJ"]
    assert header.RESERVED_INFO["CNADJ"].type == "Integer"
    assert header.RESERVED_INFO["CNADJ"].number == "."
    assert header.RESERVED_INFO["CNADJ"].description


def test_info_cicn():
    assert header.RESERVED_INFO["CICN"]
    assert header.RESERVED_INFO["CICN"].type == "Integer"
    assert header.RESERVED_INFO["CICN"].number == 2
    assert header.RESERVED_INFO["CICN"].description


def test_info_cicnadj():
    assert header.RESERVED_INFO["CICNADJ"]
    assert header.RESERVED_INFO["CICNADJ"].type == "Integer"
    assert header.RESERVED_INFO["CICNADJ"].number == "."
    assert header.RESERVED_INFO["CICNADJ"].description


# FORMAT for (small) variants -------------------------------------------------


def test_format_ad():
    assert header.RESERVED_FORMAT["AD"]
    assert header.RESERVED_FORMAT["AD"].type == "Integer"
    assert header.RESERVED_FORMAT["AD"].number == "R"
    assert header.RESERVED_FORMAT["AD"].description


def test_format_adf():
    assert header.RESERVED_FORMAT["ADF"]
    assert header.RESERVED_FORMAT["ADF"].type == "Integer"
    assert header.RESERVED_FORMAT["ADF"].number == "R"
    assert header.RESERVED_FORMAT["ADF"].description


def test_format_adr():
    assert header.RESERVED_FORMAT["ADR"]
    assert header.RESERVED_FORMAT["ADR"].type == "Integer"
    assert header.RESERVED_FORMAT["ADR"].number == "R"
    assert header.RESERVED_FORMAT["ADR"].description


def test_format_dp():
    assert header.RESERVED_FORMAT["DP"]
    assert header.RESERVED_FORMAT["DP"].type == "Integer"
    assert header.RESERVED_FORMAT["DP"].number == 1
    assert header.RESERVED_FORMAT["DP"].description


def test_format_ec():
    assert header.RESERVED_FORMAT["EC"]
    assert header.RESERVED_FORMAT["EC"].type == "Integer"
    assert header.RESERVED_FORMAT["EC"].number == "A"
    assert header.RESERVED_FORMAT["EC"].description


def test_format_ft():
    assert header.RESERVED_FORMAT["FT"]
    assert header.RESERVED_FORMAT["FT"].type == "String"
    assert header.RESERVED_FORMAT["FT"].number == "1"
    assert header.RESERVED_FORMAT["FT"].description


def test_format_gq():
    assert header.RESERVED_FORMAT["GQ"]
    assert header.RESERVED_FORMAT["GQ"].type == "Integer"
    assert header.RESERVED_FORMAT["GQ"].number == "G"
    assert header.RESERVED_FORMAT["GQ"].description


def test_format_gp():
    assert header.RESERVED_FORMAT["GP"]
    assert header.RESERVED_FORMAT["GP"].type == "Float"
    assert header.RESERVED_FORMAT["GP"].number == "G"
    assert header.RESERVED_FORMAT["GP"].description


def test_format_gt():
    assert header.RESERVED_FORMAT["GT"]
    assert header.RESERVED_FORMAT["GT"].type == "String"
    assert header.RESERVED_FORMAT["GT"].number == 1
    assert header.RESERVED_FORMAT["GT"].description


def test_format_gl():
    assert header.RESERVED_FORMAT["GL"]
    assert header.RESERVED_FORMAT["GL"].type == "Float"
    assert header.RESERVED_FORMAT["GL"].number == "G"
    assert header.RESERVED_FORMAT["GL"].description


def test_format_hq():
    assert header.RESERVED_FORMAT["HQ"]
    assert header.RESERVED_FORMAT["HQ"].type == "Integer"
    assert header.RESERVED_FORMAT["HQ"].number == 2
    assert header.RESERVED_FORMAT["HQ"].description


def test_format_mq():
    assert header.RESERVED_FORMAT["MQ"]
    assert header.RESERVED_FORMAT["MQ"].type == "Integer"
    assert header.RESERVED_FORMAT["MQ"].number == 1
    assert header.RESERVED_FORMAT["MQ"].description


def test_format_pl():
    assert header.RESERVED_FORMAT["PL"]
    assert header.RESERVED_FORMAT["PL"].type == "Integer"
    assert header.RESERVED_FORMAT["PL"].number == "G"
    assert header.RESERVED_FORMAT["PL"].description


def test_format_pq():
    assert header.RESERVED_FORMAT["PQ"]
    assert header.RESERVED_FORMAT["PQ"].type == "Integer"
    assert header.RESERVED_FORMAT["PQ"].number == 1
    assert header.RESERVED_FORMAT["PQ"].description


def test_format_ps():
    assert header.RESERVED_FORMAT["PS"]
    assert header.RESERVED_FORMAT["PS"].type == "Integer"
    assert header.RESERVED_FORMAT["PS"].number == 1
    assert header.RESERVED_FORMAT["PS"].description


# FORMAT for SVs --------------------------------------------------------------


def test_format_cn():
    assert header.RESERVED_FORMAT["CN"]
    assert header.RESERVED_FORMAT["CN"].type == "Integer"
    assert header.RESERVED_FORMAT["CN"].number == 1
    assert header.RESERVED_FORMAT["CN"].description


def test_format_cnq():
    assert header.RESERVED_FORMAT["CNQ"]
    assert header.RESERVED_FORMAT["CNQ"].type == "Float"
    assert header.RESERVED_FORMAT["CNQ"].number == 1
    assert header.RESERVED_FORMAT["CNQ"].description


def test_format_cnl():
    assert header.RESERVED_FORMAT["CNL"]
    assert header.RESERVED_FORMAT["CNL"].type == "Float"
    assert header.RESERVED_FORMAT["CNL"].number == "G"
    assert header.RESERVED_FORMAT["CNL"].description


def test_format_cnp():
    assert header.RESERVED_FORMAT["CNP"]
    assert header.RESERVED_FORMAT["CNP"].type == "Float"
    assert header.RESERVED_FORMAT["CNP"].number == "G"
    assert header.RESERVED_FORMAT["CNP"].description


def test_format_nq():
    assert header.RESERVED_FORMAT["NQ"]
    assert header.RESERVED_FORMAT["NQ"].type == "Integer"
    assert header.RESERVED_FORMAT["NQ"].number == 1
    assert header.RESERVED_FORMAT["NQ"].description


def test_format_hap():
    assert header.RESERVED_FORMAT["HAP"]
    assert header.RESERVED_FORMAT["HAP"].type == "Integer"
    assert header.RESERVED_FORMAT["HAP"].number == 1
    assert header.RESERVED_FORMAT["HAP"].description


def test_format_ahap():
    assert header.RESERVED_FORMAT["AHAP"]
    assert header.RESERVED_FORMAT["AHAP"].type == "Integer"
    assert header.RESERVED_FORMAT["AHAP"].number == 1
    assert header.RESERVED_FORMAT["AHAP"].description



# -*- coding: utf-8 -*-
"""Shared code for tests"""

import textwrap

import pytest


@pytest.fixture
def multisample_vcf():
    """Return string with multi-sample VCF"""
    return textwrap.dedent(
        """
        ##fileformat=VCFv4.3
        ##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
        ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
        #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
        20\t14370\t.\tG\tA\t29\t.\t.\tGT\t0/0\t1/0\t1/1
        20\t17330\t.\tT\tA\t3\t.\t.\tGT\t0/0\t0/1\t0/0
        20\t1110696\t.\tA\tG,T\t67\t.\t.\tGT\t1/2\t2/1\t2/2
        20\t1230237\t.\tT\t.\t47\t.\t.\tGT\t0/0\t0/0\t0/0
        20\t1234567\t.\tGTC\tG,GTCT\t50\t.\t.\tGT\t0/1\t0/2\t1/1
        """
    ).lstrip()


@pytest.fixture
def multisample_vcf_only_NA00002():
    """Return string with multi-sample VCF limited to NA00002"""
    return textwrap.dedent(
        """
        ##fileformat=VCFv4.3
        ##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
        ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
        #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00002
        20\t14370\t.\tG\tA\t29\t.\t.\tGT\t1/0
        20\t17330\t.\tT\tA\t3\t.\t.\tGT\t0/1
        20\t1110696\t.\tA\tG,T\t67\t.\t.\tGT\t2/1
        20\t1230237\t.\tT\t.\t47\t.\t.\tGT\t0/0
        20\t1234567\t.\tGTC\tG,GTCT\t50\t.\t.\tGT\t0/2
        """
    ).lstrip()


@pytest.fixture
def multisample_vcf_reordered():
    """Return string with re-ordered multi-sample VCF"""
    return textwrap.dedent(
        """
        ##fileformat=VCFv4.3
        ##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
        ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
        #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00002\tNA00003\tNA00001
        20\t14370\t.\tG\tA\t29\t.\t.\tGT\t1/0\t1/1\t0/0
        20\t17330\t.\tT\tA\t3\t.\t.\tGT\t0/1\t0/0\t0/0
        20\t1110696\t.\tA\tG,T\t67\t.\t.\tGT\t2/1\t2/2\t1/2
        20\t1230237\t.\tT\t.\t47\t.\t.\tGT\t0/0\t0/0\t0/0
        20\t1234567\t.\tGTC\tG,GTCT\t50\t.\t.\tGT\t0/2\t1/1\t0/1
        """
    ).lstrip()


@pytest.fixture
def multisample_vcf_file(tmpdir, multisample_vcf):
    """Return path to multi-sample VCF file"""
    p = tmpdir.mkdir("input").join("input.vcf")
    with open(str(p), "wt") as f:
        f.write(multisample_vcf)
    return str(p)


@pytest.fixture
def nosample_vcf():
    """Return string VCF that has no samples and no FORMAT"""
    return textwrap.dedent(
        """
        ##fileformat=VCFv4.3
        ##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
        ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
        #CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO
        20\t14370\t.\tG\tA\t29\t.\t.
        20\t17330\t.\tT\tA\t3\t.\t.
        20\t1110696\t.\tA\tG,T\t67\t.\t.
        20\t1230237\t.\tT\t.\t47\t.\t.
        20\t1234567\t.\tGTC\tG,GTCT\t50\t.\t.
        """
    ).lstrip()


@pytest.fixture
def nosample_vcf_file(tmpdir, nosample_vcf):
    """Return path to file without samples/FORMAT"""
    p = tmpdir.mkdir("input").join("input.vcf")
    with open(str(p), "wt") as f:
        f.write(nosample_vcf)
    return str(p)



# -*- coding: utf-8 -*-
"""Test the structural variant Record classes
"""

from vcfpy import record

# BreakEnd ------------------------------------------------------------------


def test_breakend_builtins():
    """Test the builtin functions of the BreakEnd class"""
    rec1 = record.BreakEnd("chr2", 1234, record.FORWARD, record.FORWARD, "A", True)
    rec2 = record.BreakEnd("chr2", 1234, record.FORWARD, record.FORWARD, "A", True)
    rec3 = record.BreakEnd("chr1", 1234, record.FORWARD, record.FORWARD, "A", True)
    assert rec1 == rec2
    assert hash(rec1) == hash(rec2)
    assert rec1 != rec3
    assert str(rec1) == "BreakEnd('chr2', 1234, '+', '+', 'A', True)"
    assert repr(rec1) == "BreakEnd('chr2', 1234, '+', '+', 'A', True)"


def test_breakend_fwd_fwd_true():
    rec = record.BreakEnd("chr2", 1234, record.FORWARD, record.FORWARD, "A", True)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '+', '+', 'A', True)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "[chr2:1234[A"
    assert EXPECTED == rec.serialize()


def test_breakend_fwd_fwd_false():
    rec = record.BreakEnd("chr2", 1234, record.FORWARD, record.FORWARD, "A", False)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '+', '+', 'A', False)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "[<chr2>:1234[A"
    assert EXPECTED == rec.serialize()


def test_breakend_fwd_rev_true():
    rec = record.BreakEnd("chr2", 1234, record.FORWARD, record.REVERSE, "A", True)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '+', '-', 'A', True)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "]chr2:1234]A"
    assert EXPECTED == rec.serialize()


def test_breakend_fwd_rev_false():
    rec = record.BreakEnd("chr2", 1234, record.FORWARD, record.REVERSE, "A", False)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '+', '-', 'A', False)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "]<chr2>:1234]A"
    assert EXPECTED == rec.serialize()


def test_breakend_rev_fwd_true():
    rec = record.BreakEnd("chr2", 1234, record.REVERSE, record.FORWARD, "A", True)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '-', '+', 'A', True)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "A[chr2:1234["
    assert EXPECTED == rec.serialize()


def test_breakend_rev_fwd_false():
    rec = record.BreakEnd("chr2", 1234, record.REVERSE, record.FORWARD, "A", False)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '-', '+', 'A', False)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "A[<chr2>:1234["
    assert EXPECTED == rec.serialize()


def test_breakend_rev_rev_true():
    rec = record.BreakEnd("chr2", 1234, record.REVERSE, record.REVERSE, "A", True)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '-', '-', 'A', True)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "A]chr2:1234]"
    assert EXPECTED == rec.serialize()


def test_breakend_rev_rev_false():
    rec = record.BreakEnd("chr2", 1234, record.REVERSE, record.REVERSE, "A", False)
    # test initialization
    EXPECTED = "BreakEnd('chr2', 1234, '-', '-', 'A', False)"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "A]<chr2>:1234]"
    assert EXPECTED == rec.serialize()


# SingleBreakEnd ------------------------------------------------------------


def test_single_breakend_fdw():
    rec = record.SingleBreakEnd(record.FORWARD, "A")
    # test initialization
    EXPECTED = "SingleBreakEnd('+', 'A')"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = ".A"
    assert EXPECTED == rec.serialize()


def test_single_breakend_rev():
    rec = record.SingleBreakEnd(record.REVERSE, "A")
    # test initialization
    EXPECTED = "SingleBreakEnd('-', 'A')"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "A."
    assert EXPECTED == rec.serialize()


# SingleBreakEnd ------------------------------------------------------------


def test_symbolic_allele():
    rec = record.SymbolicAllele("DUP")
    # test initialization
    EXPECTED = "SymbolicAllele('DUP')"
    assert EXPECTED == str(rec)
    # test serialize()
    EXPECTED = "<DUP>"
    assert EXPECTED == rec.serialize()



# -*- coding: utf-8 -*-
"""Test that the public API classes are available from the ``vcfpy`` package
"""

import io
import textwrap

import pytest

from vcfpy import exceptions
from vcfpy import header
from vcfpy import parser


# HeaderChecker ---------------------------------------------------------------


def test_header_checker_missing_fileformat():
    # construct Header
    lines = [header.HeaderLine("key", "value")]
    hdr = header.Header(lines=lines, samples=header.SamplesInfos(["NA001"]))
    # setup HeaderChecker
    stream = io.StringIO()
    checker = parser.HeaderChecker()
    # execute
    with pytest.raises(exceptions.InvalidHeaderException):
        checker.run(hdr)


def test_header_checker_unknown_vcf_version():
    # construct Header
    lines = [header.HeaderLine("fileformat", "VCFv4.4")]
    hdr = header.Header(lines=lines, samples=header.SamplesInfos(["NA001"]))
    # setup HeaderChecker
    stream = io.StringIO()
    checker = parser.HeaderChecker()
    # execute
    checker.run(hdr)
    # check result
    EXPECTED = textwrap.dedent(
        r"""
    """
    ).lstrip()
    assert stream.getvalue() == EXPECTED


def test_header_checker_known_vcf_version():
    # construct Header
    lines = [header.HeaderLine("fileformat", "VCFv4.3")]
    hdr = header.Header(lines=lines, samples=header.SamplesInfos(["NA001"]))
    # setup HeaderChecker
    stream = io.StringIO()
    checker = parser.HeaderChecker()
    # execute
    checker.run(hdr)
    # check result
    EXPECTED = ""
    assert stream.getvalue() == EXPECTED



# -*- coding: utf-8 -*-
"""Test parsing of breakend records
"""

import io
import sys

from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=ANNO,Number=.,Type=String,Description="Additional annotation">
##INFO=<ID=SVTYPE,Number=1,Type=String,Description="SV type">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FILTER=<ID=PASS,Description="All filters passed">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##FORMAT=<ID=FT,Number=1,Type=String,Description="Call-wise filters">
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


def vcf_parser(lines):
    return parser.Parser(io.StringIO(MEDIUM_HEADER + lines), "<builtin>")


def test_parse_simple_breakend():
    # Setup parser with stock header and lines to parse
    LINES = "2\t321681\tbnd_W\tG\tG]17:198982]\t6\tPASS\tSVTYPE=BND\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('2', 321681, ['bnd_W'], 'G', [BreakEnd('17', 198982, '-', '-', 'G', True)], 6, ['PASS'], "
            "OrderedDict([('SVTYPE', 'BND')]), ['GT'], [Call('NA00001', OrderedDict([('GT', '0/1')])), "
            "Call('NA00002', OrderedDict([('GT', '0/0')])), Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('2', 321681, ['bnd_W'], 'G', [BreakEnd('17', 198982, '-', '-', 'G', True)], 6, ['PASS'], "
            "{'SVTYPE': 'BND'}, ['GT'], [Call('NA00001', {'GT': '0/1'}), "
            "Call('NA00002', {'GT': '0/0'}), Call('NA00003', {'GT': '0/0'})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED
    assert RESULT.ALT[0].serialize() == "G]17:198982]"


def test_parse_breakend_with_seq():
    # Setup parser with stock header and lines to parse
    LINES = "2\t321681\tbnd_V\tT\t]13:123456]AGTNNNNNCAT\t6\tPASS\tSVTYPE=BND\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('2', 321681, ['bnd_V'], 'T', [BreakEnd('13', 123456, '+', '-', 'AGTNNNNNCAT', True)], 6, "
            "['PASS'], OrderedDict([('SVTYPE', 'BND')]), ['GT'], [Call('NA00001', OrderedDict([('GT', '0/1')])), "
            "Call('NA00002', OrderedDict([('GT', '0/0')])), Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('2', 321681, ['bnd_V'], 'T', [BreakEnd('13', 123456, '+', '-', 'AGTNNNNNCAT', True)], 6, "
            "['PASS'], {'SVTYPE': 'BND'}, ['GT'], [Call('NA00001', {'GT': '0/1'}), "
            "Call('NA00002', {'GT': '0/0'}), Call('NA00003', {'GT': '0/0'})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED
    assert RESULT.ALT[0].serialize() == "]13:123456]AGTNNNNNCAT"


def test_parse_breakend_telomere():
    # Setup parser with stock header and lines to parse
    LINES = "2\t321681\tbnd_V\tN\t.[13:123457[\t6\tPASS\tSVTYPE=BND\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('2', 321681, ['bnd_V'], 'N', [BreakEnd('13', 123457, '-', '+', '.', True)], 6, ['PASS'], "
            "OrderedDict([('SVTYPE', 'BND')]), ['GT'], [Call('NA00001', OrderedDict([('GT', '0/1')])), "
            "Call('NA00002', OrderedDict([('GT', '0/0')])), Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('2', 321681, ['bnd_V'], 'N', [BreakEnd('13', 123457, '-', '+', '.', True)], 6, ['PASS'], "
            "{'SVTYPE': 'BND'}, ['GT'], [Call('NA00001', {'GT': '0/1'}), "
            "Call('NA00002', {'GT': '0/0'}), Call('NA00003', {'GT': '0/0'})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED
    assert RESULT.ALT[0].serialize() == ".[13:123457["


def test_parse_breakend_multi_mate():
    # Setup parser with stock header and lines to parse
    LINES = "2\t321681\tbnd_U\tT\tC[2:321682[,C[17:198983\t6\tPASS\tSVTYPE=BND\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('2', 321681, ['bnd_U'], 'T', [BreakEnd('2', 321682, '-', '+', 'C', True), "
            "BreakEnd('17', 198983, '-', '+', 'C', True)], 6, ['PASS'], OrderedDict([('SVTYPE', 'BND')]), ['GT'], "
            "[Call('NA00001', OrderedDict([('GT', '0/1')])), Call('NA00002', OrderedDict([('GT', '0/0')])), "
            "Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('2', 321681, ['bnd_U'], 'T', [BreakEnd('2', 321682, '-', '+', 'C', True), "
            "BreakEnd('17', 198983, '-', '+', 'C', True)], 6, ['PASS'], {'SVTYPE': 'BND'}, ['GT'], "
            "[Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/0'}), Call('NA00003', {'GT': '0/0'})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED
    assert RESULT.ALT[0].serialize() == "C[2:321682["
    assert RESULT.ALT[1].serialize() == "C[17:198983["


def test_parse_breakend_single_breakend_fwd():
    # Setup parser with stock header and lines to parse
    LINES = "13\t123457\tbnd_X\tA\t.A\t6\tPASS\tSVTYPE=BND\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('13', 123457, ['bnd_X'], 'A', [SingleBreakEnd('+', 'A')], 6, ['PASS'], "
            "OrderedDict([('SVTYPE', 'BND')]), ['GT'], [Call('NA00001', OrderedDict([('GT', '0/1')])), "
            "Call('NA00002', OrderedDict([('GT', '0/0')])), Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('13', 123457, ['bnd_X'], 'A', [SingleBreakEnd('+', 'A')], 6, ['PASS'], "
            "{'SVTYPE': 'BND'}, ['GT'], [Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/0'}), "
            "Call('NA00003', {'GT': '0/0'})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED
    assert RESULT.ALT[0].serialize() == ".A"


def test_parse_breakend_single_breakend_rev():
    # Setup parser with stock header and lines to parse
    LINES = "13\t123457\tbnd_X\tA\tA.\t6\tPASS\tSVTYPE=BND\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('13', 123457, ['bnd_X'], 'A', [SingleBreakEnd('-', 'A')], 6, ['PASS'], "
            "OrderedDict([('SVTYPE', 'BND')]), ['GT'], [Call('NA00001', OrderedDict([('GT', '0/1')])), "
            "Call('NA00002', OrderedDict([('GT', '0/0')])), "
            "Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('13', 123457, ['bnd_X'], 'A', [SingleBreakEnd('-', 'A')], 6, ['PASS'], "
            "{'SVTYPE': 'BND'}, ['GT'], [Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/0'}), "
            "Call('NA00003', {'GT': '0/0'})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED
    assert RESULT.ALT[0].serialize() == "A."



# -*- coding: utf-8 -*-
"""Tests for reading with file without samples"""

from io import StringIO
import textwrap

import vcfpy


def test_reading_parse_nosample(tmpdir, nosample_vcf_file):
    """Test reading of VCF file without any samples."""
    # Perform record-wise copying, saving results in records
    records = []
    with vcfpy.Reader.from_path(nosample_vcf_file) as reader:
        for record in reader:
            records.append(record)
    # Check resulting records, checking the first and last records is enough
    assert len(records) == 5


def test_reading_parse_nosample(tmpdir, nosample_vcf_file):
    """Read VCF file without samples, write file with samples."""
    # Perform record-wise copying, saving results in records
    path_out = tmpdir.mkdir("output").join("output.vcf")
    with vcfpy.Reader.from_path(nosample_vcf_file) as reader:
        header = reader.header.copy()
        header.samples = vcfpy.SamplesInfos(["NA00001", "NA00002", "NA00003"])
        with vcfpy.Writer.from_path(str(path_out), header) as writer:
            for record in reader:
                record.update_calls(
                    [vcfpy.Call(sample, {}) for sample in ("NA00001", "NA00002", "NA00003")]
                )
                record.add_format("GT", "./.")
                writer.write_record(record)

    expected = textwrap.dedent(
        """
    ##fileformat=VCFv4.3
    ##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
    ##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
    #CHROM	POS	ID	REF	ALT	QUAL	FILTER	INFO	FORMAT	NA00001	NA00002	NA00003
    20	14370	.	G	A	29	.	.	GT	.	.	.
    20	17330	.	T	A	3	.	.	GT	.	.	.
    20	1110696	.	A	G,T	67	.	.	GT	.	.	.
    20	1230237	.	T	.	47	.	.	GT	.	.	.
    20	1234567	.	GTC	G,GTCT	50	.	.	GT	.	.	.
    """
    ).lstrip()

    assert path_out.open("rt").read() == expected



# -*- coding: utf-8 -*-
"""Code for representing a VCF record

The VCF record structure is modeled after the one of PyVCF
"""

import re


#: Code for single nucleotide variant allele
SNV = "SNV"
#: Code for a multi nucleotide variant allele
MNV = "MNV"
#: Code for "clean" deletion allele
DEL = "DEL"
#: Code for "clean" insertion allele
INS = "INS"
#: Code for indel allele, includes substitutions of unequal length
INDEL = "INDEL"
#: Code for structural variant allele
SV = "SV"
#: Code for break-end allele
BND = "BND"
#: Code for symbolic allele that is neither SV nor BND
SYMBOLIC = "SYMBOLIC"

#: Code for mixed variant type
MIXED = "MIXED"

#: Code for homozygous reference
HOM_REF = 0
#: Code for heterozygous
HET = 1
#: Code for homozygous alternative
HOM_ALT = 2

#: Characters reserved in VCF, have to be escaped in INFO fields
RESERVED_CHARS = {"INFO": ";%,\r\n\t", "FORMAT": ":=%,\r\n\t"}
#: Mapping for escaping reserved characters
ESCAPE_MAPPING = [
    ("%", "%25"),
    (":", "%3A"),
    (";", "%3B"),
    ("=", "%3D"),
    (",", "%2C"),
    ("\r", "%0D"),
    ("\n", "%0A"),
    ("\t", "%09"),
]
#: Mapping from escaped characters to reserved one
UNESCAPE_MAPPING = [(v, k) for k, v in ESCAPE_MAPPING]


class Record:
    """Represent one record from the VCF file

    Record objects are iterators of their calls
    """

    def __init__(self, CHROM, POS, ID, REF, ALT, QUAL, FILTER, INFO, FORMAT=None, calls=None):
        if bool(FORMAT) != bool(calls):
            raise ValueError("Either provide both FORMAT and calls or none.")
        #: A ``str`` with the chromosome name
        self.CHROM = CHROM
        #: An ``int`` with a 1-based begin position
        self.POS = POS
        #: An ``int`` with a 0-based begin position
        self.begin = POS - 1
        #: An ``int`` with a 0-based end position
        self.end = None  # XXX
        #: A list of the semicolon-separated values of the ID column
        self.ID = list(ID)
        #: A ``str`` with the REF value
        self.REF = REF
        #: A list of alternative allele records of type :py:class:`AltRecord`
        self.ALT = list(ALT)
        #: The quality value, can be ``None``
        self.QUAL = QUAL
        #: A list of strings for the FILTER column
        self.FILTER = FILTER
        #: An OrderedDict giving the values of the INFO column, flags are
        #: mapped to ``True``
        self.INFO = INFO
        #: A list of strings for the FORMAT column.  Optional, must be given if
        #: and only if ``calls`` is also given.
        self.FORMAT = FORMAT or []
        #: A list of genotype :py:class:`Call` objects.  Optional, must be given if
        #: and only if ``FORMAT`` is also given.
        self.calls = calls or []
        #: A mapping from sample name to entry in self.calls.
        self.call_for_sample = {}
        self.update_calls(self.calls)

    def update_calls(self, calls):
        """Update ``self.calls`` and other fields as necessary."""
        for call in calls:
            call.site = self
        self.call_for_sample = {call.sample: call for call in calls}

    def is_snv(self):
        """Return ``True`` if it is a SNV"""
        return len(self.REF) == 1 and all(a.type == "SNV" for a in self.ALT)

    @property
    def affected_start(self):
        """Return affected start position in 0-based coordinates

        For SNVs, MNVs, and deletions, the behaviour is the start position.
        In the case of insertions, the position behind the insert position is
        returned, yielding a 0-length interval together with
        :py:meth:`~Record.affected_end`
        """
        types = {alt.type for alt in self.ALT}  # set!
        BAD_MIX = {INS, SV, BND, SYMBOLIC}  # don't mix well with others
        if (BAD_MIX & types) and len(types) == 1 and list(types)[0] == INS:
            # Only insertions, return 0-based position right of first base
            return self.POS  # right of first base
        else:  # Return 0-based start position of first REF base
            return self.POS - 1  # left of first base

    @property
    def affected_end(self):
        """Return affected start position in 0-based coordinates

        For SNVs, MNVs, and deletions, the behaviour is based on the start
        position and the length of the REF.  In the case of insertions, the
        position behind the insert position is returned, yielding a 0-length
        interval together with :py:meth:`~Record.affected_start`
        """
        types = {alt.type for alt in self.ALT}  # set!
        BAD_MIX = {INS, SV, BND, SYMBOLIC}  # don't mix well with others
        if (BAD_MIX & types) and len(types) == 1 and list(types)[0] == INS:
            # Only insertions, return 0-based position right of first base
            return self.POS  # right of first base
        else:  # Return 0-based end position, behind last REF base
            return (self.POS - 1) + len(self.REF)

    def add_filter(self, label):
        """Add label to FILTER if not set yet, removing ``PASS`` entry if
        present
        """
        if label not in self.FILTER:
            if "PASS" in self.FILTER:
                self.FILTER = [f for f in self.FILTER if f != "PASS"]
            self.FILTER.append(label)

    def add_format(self, key, value=None):
        """Add an entry to format

        The record's calls ``data[key]`` will be set to ``value`` if not yet
        set and value is not ``None``.  If key is already in FORMAT then
        nothing is done.
        """
        if key in self.FORMAT:
            return
        self.FORMAT.append(key)
        if value is not None:
            for call in self:
                call.data.setdefault(key, value)

    def __iter__(self):
        """Return generator yielding from ``self.calls``"""
        yield from self.calls

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __str__(self):
        tpl = "Record({})"
        lst = [
            self.CHROM,
            self.POS,
            self.ID,
            self.REF,
            self.ALT,
            self.QUAL,
            self.FILTER,
            self.INFO,
            self.FORMAT,
            self.calls,
        ]
        return tpl.format(", ".join(map(repr, lst)))

    def __repr__(self):
        return str(self)


class UnparsedCall:
    """Placeholder for :py:class:`Call` when parsing only a subset of fields"""

    def __init__(self, sample, unparsed_data, site=None):
        #: the name of the sample for which the call was made
        self.sample = sample
        #: ``str`` with the unparsed data
        self.unparsed_data = unparsed_data
        #: the :py:class:`Record` of this :py:class:`Call`
        self.site = site


#: Regular expression for splitting alleles
ALLELE_DELIM = re.compile(r"[|/]")


class Call:
    """The information for a genotype callable

    By VCF, this should always include the genotype information and
    can contain an arbitrary number of further annotation, e.g., the
    coverage at the variant position.
    """

    def __init__(self, sample, data, site=None):
        #: the name of the sample for which the call was made
        self.sample = sample
        #: an OrderedDict with the key/value pair information from the
        #: call's data
        self.data = data
        #: the :py:class:`Record` of this :py:class:`Call`
        self.site = site
        #: the allele numbers (0, 1, ...) in this calls or None for no-call
        self.gt_alleles = None
        #: whether or not the variant is fully called
        self.called = None
        #: the number of alleles in this sample's call
        self.ploidy = None
        self._genotype_updated()

    def set_genotype(self, genotype):
        """Set ``self.data["GT"]`` to ``genotype`` and properly update related
        properties.
        """
        self.data["GT"] = genotype
        self._genotype_updated()

    def _genotype_updated(self):
        """Update fields related to ``self.data["GT"]``."""
        if self.data.get("GT", None) is None:
            self.gt_alleles = None
            self.called = None
            self.ploidy = None
        else:
            self.gt_alleles = []
            for allele in ALLELE_DELIM.split(str(self.data["GT"])):
                if allele == ".":
                    self.gt_alleles.append(None)
                else:
                    self.gt_alleles.append(int(allele))
            self.called = all([al is not None for al in self.gt_alleles])
            self.ploidy = len(self.gt_alleles)

    @property
    def is_phased(self):
        """Return boolean indicating whether this call is phased"""
        return "|" in self.data.get("GT", "")

    @property
    def gt_phase_char(self):
        """Return character to use for phasing"""
        return "/" if not self.is_phased else "|"

    @property
    def gt_bases(self):
        """Return the actual genotype bases, e.g. if VCF genotype is 0/1,
        could return ('A', 'T')
        """
        result = []
        for a in self.gt_alleles:
            if a is None:
                result.append(None)
            elif a == 0:
                result.append(self.site.REF)
            else:
                result.append(self.site.ALT[a - 1].value)
        return tuple(result)

    @property
    def gt_type(self):
        """The type of genotype, returns one of ``HOM_REF``, ``HOM_ALT``, and
        ``HET``.
        """
        if not self.called:
            return None  # not called
        elif all(a == 0 for a in self.gt_alleles):
            return HOM_REF
        elif len(set(self.gt_alleles)) == 1:
            return HOM_ALT
        else:
            return HET

    def is_filtered(self, require=None, ignore=None):
        """Return ``True`` for filtered calls

        :param iterable ignore: if set, the filters to ignore, make sure to
            include 'PASS', when setting, default is ``['PASS']``
        :param iterable require: if set, the filters to require for returning
            ``True``
        """
        ignore = ignore or ["PASS"]
        if "FT" not in self.data or not self.data["FT"]:
            return False
        for ft in self.data["FT"]:
            if ft in ignore:
                continue  # skip
            if not require:
                return True
            elif ft in require:
                return True
        return False

    @property
    def is_het(self):
        """Return ``True`` for heterozygous calls"""
        return self.gt_type == HET

    @property
    def is_variant(self):
        """Return ``True`` for non-hom-ref calls"""
        return bool(self.gt_type)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __str__(self):
        tpl = "Call({})"
        lst = [self.sample, self.data]
        return tpl.format(", ".join(map(repr, lst)))

    def __repr__(self):
        return str(self)


class AltRecord:
    """An alternative allele Record

    Currently, can be a substitution, an SV placeholder, or breakend
    """

    def __init__(self, type_=None):
        #: String describing the type of the variant, could be one of
        #: SNV, MNV, could be any of teh types described in the ALT
        #: header lines, such as DUP, DEL, INS, ...
        self.type = type_

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def serialize(self):
        """Return ``str`` with representation for VCF file"""
        raise NotImplementedError("Abstract class, implemented in sub class")


class Substitution(AltRecord):
    """A basic alternative allele record describing a REF->AltRecord
    substitution

    Note that this subsumes MNVs, insertions, and deletions.
    """

    def __init__(self, type_, value):
        super().__init__(type_)
        #: The alternative base sequence to use in the substitution
        self.value = value

    def serialize(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __str__(self):
        tpl = "Substitution(type_={}, value={})"
        return tpl.format(*map(repr, [self.type, self.value]))

    def __repr__(self):
        return str(self)


#: code for five prime orientation :py:class:`BreakEnd`
FIVE_PRIME = "5"
#: code for three prime orientation :py:class:`BreakEnd`
THREE_PRIME = "3"

#: code for forward orientation
FORWARD = "+"
#: code for reverse orientation
REVERSE = "-"


class BreakEnd(AltRecord):
    """A placeholder for a breakend"""

    def __init__(
        self, mate_chrom, mate_pos, orientation, mate_orientation, sequence, within_main_assembly
    ):
        super().__init__("BND")
        #: chromosome of the mate breakend
        self.mate_chrom = mate_chrom
        #: position of the mate breakend
        self.mate_pos = mate_pos
        #: orientation of this breakend
        self.orientation = orientation
        #: orientation breakend's mate
        self.mate_orientation = mate_orientation
        #: breakpoint's connecting sequence
        self.sequence = sequence
        #: ``bool`` specifying if the breakend mate is within the assembly
        #: (``True``) or in an ancillary assembly (``False``)
        self.within_main_assembly = within_main_assembly

    def serialize(self):
        """Return string representation for VCF"""
        if self.mate_chrom is None:
            remote_tag = "."
        else:
            if self.within_main_assembly:
                mate_chrom = self.mate_chrom
            else:
                mate_chrom = "<{}>".format(self.mate_chrom)
            tpl = {FORWARD: "[{}:{}[", REVERSE: "]{}:{}]"}[self.mate_orientation]
            remote_tag = tpl.format(mate_chrom, self.mate_pos)
        if self.orientation == FORWARD:
            return remote_tag + self.sequence
        else:
            return self.sequence + remote_tag

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __str__(self):
        tpl = "BreakEnd({})"
        vals = [
            self.mate_chrom,
            self.mate_pos,
            self.orientation,
            self.mate_orientation,
            self.sequence,
            self.within_main_assembly,
        ]
        return tpl.format(", ".join(map(repr, vals)))

    def __repr__(self):
        return str(self)


class SingleBreakEnd(BreakEnd):
    """A placeholder for a single breakend"""

    def __init__(self, orientation, sequence):
        super().__init__(None, None, orientation, None, sequence, None)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __str__(self):
        tpl = "SingleBreakEnd({})"
        vals = [self.orientation, self.sequence]
        return tpl.format(", ".join(map(repr, vals)))


class SymbolicAllele(AltRecord):
    """A placeholder for a symbolic allele

    The allele symbol must be defined in the header using an ``ALT`` header
    before being parsed.  Usually, this is used for succinct descriptions of
    structural variants or IUPAC parameters.
    """

    def __init__(self, value):
        super().__init__(SYMBOLIC)
        #: The symbolic value, e.g. 'DUP'
        self.value = value

    def serialize(self):
        return "<{}>".format(self.value)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def __str__(self):
        return "SymbolicAllele({})".format(repr(self.value))

    def __repr__(self):
        return str(self)



# -*- coding: utf-8 -*-
"""Roundtrip-test: read in VCF file and write out again
"""

import os

from vcfpy import reader
from vcfpy import writer

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


def test_vcf_roundtrip(tmpdir_factory):
    # open reader with VCF file to read from
    in_path = os.path.join(os.path.dirname(__file__), "vcfs/full_vcf43.vcf")
    r = reader.Reader.from_path(in_path)
    # open temporary file and setup the Writer with header info from reader
    out_path = tmpdir_factory.mktemp("write_header").join("out.vcf")
    w = writer.Writer.from_path(out_path, r.header)
    # copy records to output file
    for rec in r:
        w.write_record(rec)
    r.close()
    w.close()
    # compare actual result with expected
    RESULT = out_path.read()
    EXPECTED = open(in_path, "rt").read()
    assert EXPECTED == RESULT



# -*- coding: utf-8 -*-
"""Test that the public API classes are available from the ``vcfpy`` package
"""

import vcfpy


def test_from_exceptions():
    assert vcfpy.VCFPyException
    assert vcfpy.InvalidHeaderException
    assert vcfpy.InvalidRecordException
    assert vcfpy.IncorrectVCFFormat
    assert vcfpy.HeaderNotFound


def test_from_header():
    assert vcfpy.Header
    assert vcfpy.HeaderLine
    assert vcfpy.SimpleHeaderLine
    assert vcfpy.AltAlleleHeaderLine
    assert vcfpy.ContigHeaderLine
    assert vcfpy.FilterHeaderLine
    assert vcfpy.MetaHeaderLine
    assert vcfpy.PedigreeHeaderLine
    assert vcfpy.SampleHeaderLine
    assert vcfpy.CompoundHeaderLine
    assert vcfpy.InfoHeaderLine
    assert vcfpy.FormatHeaderLine
    assert vcfpy.SamplesInfos


def test_from_record():
    assert vcfpy.Record
    assert vcfpy.UnparsedCall
    assert vcfpy.Call
    assert vcfpy.AltRecord
    assert vcfpy.Substitution
    assert vcfpy.SV
    assert vcfpy.BreakEnd
    assert vcfpy.SymbolicAllele


def test_record_constants():
    assert vcfpy.SNV
    assert vcfpy.MNV
    assert vcfpy.DEL
    assert vcfpy.INS
    assert vcfpy.INDEL
    assert vcfpy.SV
    assert vcfpy.BND
    assert vcfpy.SYMBOLIC
    assert vcfpy.MIXED

    assert vcfpy.HOM_REF is not None
    assert vcfpy.HET
    assert vcfpy.HOM_ALT


def test_from_reader():
    assert vcfpy.Reader


def test_from_writer():
    assert vcfpy.Writer



# -*- coding: utf-8 -*-
"""Tests for reading with parsing only a subset of samples"""

from vcfpy import Reader, Writer, Call, UnparsedCall


def test_reading_parse_subset(tmpdir, multisample_vcf_file):
    # Perform record-wise copying, saving results in records
    records = []
    out_path = str(tmpdir.mkdir("output").join("output.vcf"))
    with Reader.from_path(multisample_vcf_file, parsed_samples=["NA00001"]) as reader:
        with Writer.from_path(out_path, reader.header) as writer:
            for record in reader:
                records.append(record)
                writer.write_record(record)
    # Check resulting records, checking the first and last records is enough
    assert len(records) == 5
    assert set(records[0].call_for_sample.keys()) == {"NA00001", "NA00002", "NA00003"}
    assert set(records[-1].call_for_sample.keys()) == {"NA00001", "NA00002", "NA00003"}
    assert isinstance(records[0].call_for_sample["NA00001"], Call)
    assert isinstance(records[0].call_for_sample["NA00002"], UnparsedCall)
    assert isinstance(records[0].call_for_sample["NA00003"], UnparsedCall)
    assert isinstance(records[-1].call_for_sample["NA00001"], Call)
    assert isinstance(records[-1].call_for_sample["NA00002"], UnparsedCall)
    assert isinstance(records[-1].call_for_sample["NA00003"], UnparsedCall)
    # Check resulting file
    with open(multisample_vcf_file, "rt") as inf, open(out_path, "rt") as outf:
        assert inf.read() == outf.read()



# -*- coding: utf-8 -*-
"""Test parsing of VCF header lines from strings
"""

import sys

import pytest

from vcfpy import header
from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


# parser.StupidHeaderLineParser.parse_key_value() --------------------------


def test_stupid_vcf_header_line_parser_file_format():
    p = parser.StupidHeaderLineParser()
    INPUT = ("fileFormat", "VCFv4.2")
    EXPECTED = "HeaderLine('fileFormat', 'VCFv4.2')"
    assert str(p.parse_key_value(*INPUT)) == EXPECTED


# parser.MappingHeaderLineParser.parse_key_value() -------------------------


def test_mapping_vcf_header_line_parser_parse_key_value_filter():
    p = parser.MappingHeaderLineParser(header.FilterHeaderLine)
    INPUT = ("FILTER", '<ID=q10,Description="Quality below 10">')
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FilterHeaderLine('FILTER', '<ID=q10,Description=\"Quality below 10\">', "
            "OrderedDict([('ID', 'q10'), ('Description', 'Quality below 10')]))"
        )
    else:
        EXPECTED = (
            "FilterHeaderLine('FILTER', '<ID=q10,Description=\"Quality below 10\">', "
            "{'ID': 'q10', 'Description': 'Quality below 10'})"
        )
    assert str(p.parse_key_value(*INPUT)) == EXPECTED


def test_mapping_vcf_header_line_parser_parse_key_value_format():
    p = parser.MappingHeaderLineParser(header.FormatHeaderLine)
    INPUT = ("FORMAT", '<ID=GT,Number=1,Type=String,Description="Genotype">')
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=GT,Number=1,Type=String,Description=\"Genotype\">', "
            "OrderedDict([('ID', 'GT'), ('Number', 1), ('Type', 'String'), ('Description', 'Genotype')]))"
        )
    else:
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=GT,Number=1,Type=String,Description=\"Genotype\">', "
            "{'ID': 'GT', 'Number': 1, 'Type': 'String', 'Description': 'Genotype'})"
        )
    assert str(p.parse_key_value(*INPUT)) == EXPECTED


def test_mapping_vcf_header_line_parser_parse_key_value_info():
    p = parser.MappingHeaderLineParser(header.InfoHeaderLine)
    INPUT = ("INFO", "<ID=NS,Number=1,Type=Integer,Description=" '"Number of Samples With Data">')
    if sys.version_info < (3, 6):
        EXPECTED = (
            "InfoHeaderLine('INFO', '<ID=NS,Number=1,Type=Integer,Description=\"Number of Samples With Data\">', "
            "OrderedDict([('ID', 'NS'), ('Number', 1), ('Type', 'Integer'), "
            "('Description', 'Number of Samples With Data')]))"
        )
    else:
        EXPECTED = (
            "InfoHeaderLine('INFO', '<ID=NS,Number=1,Type=Integer,Description=\"Number of Samples With Data\">', "
            "{'ID': 'NS', 'Number': 1, 'Type': 'Integer', 'Description': 'Number of Samples With Data'})"
        )
    assert str(p.parse_key_value(*INPUT)) == EXPECTED


def test_mapping_vcf_header_line_parser_parse_key_value_contig():
    p = parser.MappingHeaderLineParser(header.ContigHeaderLine)
    INPUT = (
        "contig",
        '<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>',
    )
    if sys.version_info < (3, 6):
        EXPECTED = (
            "ContigHeaderLine('contig', '<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,"
            "species=\"Homo sapiens\",taxonomy=x>', OrderedDict([('ID', '20'), ('length', '62435964'), "
            "('assembly', 'B36'), ('md5', 'f126cdf8a6e0c7f379d618ff66beb2da'), ('species', 'Homo sapiens'), "
            "('taxonomy', 'x')]))"
        )
    else:
        EXPECTED = (
            "ContigHeaderLine('contig', '<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,"
            "species=\"Homo sapiens\",taxonomy=x>', {'ID': '20', 'length': '62435964', 'assembly': 'B36', "
            "'md5': 'f126cdf8a6e0c7f379d618ff66beb2da', 'species': 'Homo sapiens', 'taxonomy': 'x'})"
        )
    assert str(p.parse_key_value(*INPUT)) == EXPECTED


# parser.HeaderParser.parse_line() -----------------------------------------


def test_vcf_header_parser_file_format():
    p = parser.HeaderParser()
    INPUT = "##fileFormat=VCFv4.2\n"
    EXPECTED = "HeaderLine('fileFormat', 'VCFv4.2')"
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_vcf_header_parser_parse_line_filter():
    p = parser.HeaderParser()
    INPUT = '##FILTER=<ID=q10,Description="Quality below 10">\n'
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FilterHeaderLine('FILTER', '<ID=q10,Description=\"Quality below 10\">', OrderedDict([('ID', 'q10'), "
            "('Description', 'Quality below 10')]))"
        )
    else:
        EXPECTED = (
            "FilterHeaderLine('FILTER', '<ID=q10,Description=\"Quality below 10\">', "
            "{'ID': 'q10', 'Description': 'Quality below 10'})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_mapping_vcf_header_parser_parse_line_format():
    p = parser.HeaderParser()
    INPUT = '##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">\n'
    if sys.version_info < (3, 6):
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=GT,Number=1,Type=String,Description=\"Genotype\">', "
            "OrderedDict([('ID', 'GT'), ('Number', 1), ('Type', 'String'), ('Description', 'Genotype')]))"
        )
    else:
        EXPECTED = (
            "FormatHeaderLine('FORMAT', '<ID=GT,Number=1,Type=String,Description=\"Genotype\">', "
            "{'ID': 'GT', 'Number': 1, 'Type': 'String', 'Description': 'Genotype'})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_mapping_vcf_header_parser_parse_line_info():
    p = parser.HeaderParser()
    INPUT = "##INFO=" "<ID=NS,Number=1,Type=Integer,Description=" '"Number of Samples With Data">\n'
    if sys.version_info < (3, 6):
        EXPECTED = (
            "InfoHeaderLine('INFO', '<ID=NS,Number=1,Type=Integer,Description=\"Number of Samples With Data\">', "
            "OrderedDict([('ID', 'NS'), ('Number', 1), ('Type', 'Integer'), ('Description', "
            "'Number of Samples With Data')]))"
        )
    else:
        EXPECTED = (
            "InfoHeaderLine('INFO', '<ID=NS,Number=1,Type=Integer,Description=\"Number of Samples With Data\">', "
            "{'ID': 'NS', 'Number': 1, 'Type': 'Integer', 'Description': 'Number of Samples With Data'})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_mapping_vcf_header_parser_parse_line_contig():
    p = parser.HeaderParser()
    INPUT = (
        "##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,"
        'species="Homo sapiens",taxonomy=x>\n'
    )
    if sys.version_info < (3, 6):
        EXPECTED = (
            "ContigHeaderLine('contig', '<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,"
            "species=\"Homo sapiens\",taxonomy=x>', OrderedDict([('ID', '20'), ('length', '62435964'), "
            "('assembly', 'B36'), ('md5', 'f126cdf8a6e0c7f379d618ff66beb2da'), ('species', 'Homo sapiens'), "
            "('taxonomy', 'x')]))"
        )
    else:
        EXPECTED = (
            "ContigHeaderLine('contig', '<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,"
            "species=\"Homo sapiens\",taxonomy=x>', {'ID': '20', 'length': '62435964', 'assembly': 'B36', "
            "'md5': 'f126cdf8a6e0c7f379d618ff66beb2da', 'species': 'Homo sapiens', 'taxonomy': 'x'})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_mapping_vcf_header_parser_parse_line_alt_allele():
    p = parser.HeaderParser()
    INPUT = "##ALT=" '<ID=R,Description="IUPAC code R = A/G">\n'
    if sys.version_info < (3, 6):
        EXPECTED = (
            "AltAlleleHeaderLine('ALT', "
            "'<ID=R,Description=\"IUPAC code R = A/G\">', "
            "OrderedDict([('ID', 'R'), "
            "('Description', 'IUPAC code R = A/G')]))"
        )
    else:
        EXPECTED = (
            "AltAlleleHeaderLine('ALT', "
            "'<ID=R,Description=\"IUPAC code R = A/G\">', "
            "{'ID': 'R', 'Description': 'IUPAC code R = A/G'})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_mapping_vcf_header_parser_parse_line_meta():
    p = parser.HeaderParser()
    INPUT = "##META=" "<ID=Assay,Type=String,Number=.,Values=[WholeGenome, Exome]>\n"
    if sys.version_info < (3, 6):
        EXPECTED = (
            "MetaHeaderLine('META', '<ID=Assay,Type=String,Number=.,"
            "Values=[WholeGenome, Exome]>', OrderedDict([('ID', 'Assay'), "
            "('Type', 'String'), ('Number', '.'), ('Values', ['WholeGenome', "
            "'Exome'])]))"
        )
    else:
        EXPECTED = (
            "MetaHeaderLine('META', '<ID=Assay,Type=String,Number=.,Values=[WholeGenome, Exome]>', "
            "{'ID': 'Assay', 'Type': 'String', 'Number': '.', 'Values': ['WholeGenome', 'Exome']})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_mapping_vcf_header_parser_parse_line_pedigree():
    p = parser.HeaderParser()
    INPUT = "##PEDIGREE=" "<ID=TumourSample,Original=GermlineID>\n"
    if sys.version_info < (3, 6):
        EXPECTED = (
            "PedigreeHeaderLine('PEDIGREE', "
            "'<ID=TumourSample,Original=GermlineID>',"
            " OrderedDict([('ID', 'TumourSample'), "
            "('Original', 'GermlineID')]))"
        )
    else:
        EXPECTED = (
            "PedigreeHeaderLine('PEDIGREE', '<ID=TumourSample,Original=GermlineID>', "
            "{'ID': 'TumourSample', 'Original': 'GermlineID'})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED


def test_mapping_vcf_header_parser_parse_line_sample():
    p = parser.HeaderParser()
    INPUT = (
        "##SAMPLE="
        "<ID=Sample1,Assay=WholeGenome,Ethnicity=AFR,Disease=None,"
        'Description="Patient germline genome from unaffected",'
        "DOI=url>\n"
    )
    if sys.version_info < (3, 6):
        EXPECTED = (
            "SampleHeaderLine('SAMPLE', '<ID=Sample1,Assay=WholeGenome,"
            'Ethnicity=AFR,Disease=None,Description="Patient germline genome from '
            "unaffected\",DOI=url>', OrderedDict([('ID', 'Sample1'), ('Assay', "
            "'WholeGenome'), ('Ethnicity', 'AFR'), ('Disease', 'None'), "
            "('Description', 'Patient germline genome from unaffected'), "
            "('DOI', 'url')]))"
        )
    else:
        EXPECTED = (
            "SampleHeaderLine('SAMPLE', '<ID=Sample1,Assay=WholeGenome,"
            'Ethnicity=AFR,Disease=None,Description="Patient germline genome from '
            "unaffected\",DOI=url>', {'ID': 'Sample1', 'Assay': 'WholeGenome', 'Ethnicity': 'AFR', 'Disease': 'None', "
            "'Description': 'Patient germline genome from unaffected', 'DOI': 'url'})"
        )
    assert str(p.parse_line(INPUT)) == EXPECTED



# -*- coding: utf-8 -*-
"""Test parsing of symbolic records
"""

import io
import sys

from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=ANNO,Number=.,Type=String,Description="Additional annotation">
##INFO=<ID=SVTYPE,Number=1,Type=String,Description="SV type">
##INFO=<ID=END,Number=1,Type=Integer,Description="SV end position">
##INFO=<ID=SVLEN,Number=1,Type=Integer,Description="SV length">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FILTER=<ID=PASS,Description="All filters passed">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##FORMAT=<ID=FT,Number=1,Type=String,Description="Call-wise filters">
##ALT=<ID=DUP,Description="Duplication">
##ALT=<ID=R,Description="IUPAC code R = A/G">
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


def vcf_parser(lines):
    return parser.Parser(io.StringIO(MEDIUM_HEADER + lines), "<builtin>")


def test_parse_dup():
    # Setup parser with stock header and lines to parse
    LINES = "2\t321681\t.\tN\t<DUP>\t.\tPASS\tSVTYPE=DUP;END=324681;SVLEN=3000\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('2', 321681, [], 'N', [SymbolicAllele('DUP')], None, ['PASS'], "
            "OrderedDict([('SVTYPE', 'DUP'), ('END', 324681), ('SVLEN', 3000)]), ['GT'], ["
            "Call('NA00001', OrderedDict([('GT', '0/1')])), Call('NA00002', OrderedDict([('GT', '0/0')])), "
            "Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('2', 321681, [], 'N', [SymbolicAllele('DUP')], None, ['PASS'], "
            "{'SVTYPE': 'DUP', 'END': 324681, 'SVLEN': 3000}, ['GT'], ["
            "Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/0'}), Call('NA00003', {'GT': '0/0'})])"
        )
    rec = p.parse_next_record()
    assert str(rec) == EXPECTED
    assert rec.ALT[0].serialize() == "<DUP>"


def test_parse_iupac():
    # Setup parser with stock header and lines to parse
    LINES = "2\t321681\t.\tC\t<R>\t.\tPASS\t.\tGT\t0/1\t0/0\t0/0\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('2', 321681, [], 'C', [SymbolicAllele('R')], None, ['PASS'], OrderedDict(), ['GT'], "
            "[Call('NA00001', OrderedDict([('GT', '0/1')])), Call('NA00002', OrderedDict([('GT', '0/0')])), "
            "Call('NA00003', OrderedDict([('GT', '0/0')]))])"
        )
    else:
        EXPECTED = (
            "Record('2', 321681, [], 'C', [SymbolicAllele('R')], None, ['PASS'], {}, ['GT'], "
            "[Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/0'}), Call('NA00003', {'GT': '0/0'})])"
        )
    rec = p.parse_next_record()
    assert str(rec) == EXPECTED
    assert rec.ALT[0].serialize() == "<R>"



# -*- coding: utf-8 -*-
"""Test writing of BGZF files
"""

import codecs
import gzip
import io

import pytest

import vcfpy
from vcfpy import parser
from vcfpy import writer
from vcfpy import record

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=ANNO,Number=.,Type=String,Description="Additional annotation">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##FORMAT=<ID=FT,Number=1,Type=String,Description="Call-wise filters">
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


@pytest.fixture(scope="function")
def header_samples():
    p = parser.Parser(stream=io.StringIO(MEDIUM_HEADER), path="<builtin>")
    p.parse_header()
    return (p.header, p.samples)


def check_file(path, line):
    """Test whether the file is the expected ``.vcf.gz`` file"""
    raw = path.read(mode="rb")
    assert raw[0] == 0x1F
    assert raw[1] == 0x8B
    # compare actual result with expected
    inflated = gzip.decompress(raw)
    RESULT = codecs.latin_1_decode(inflated)[0]
    LINE = "20\t100\t.\tC\tT\t.\t.\t.\tGT\t0/1\t0/0\t1/1\n"
    EXPECTED = MEDIUM_HEADER + LINE
    assert EXPECTED == RESULT


def test_write_minimal_record_writer_from_path(header_samples, tmpdir_factory):
    O = vcfpy.OrderedDict
    # open temporary file and setup the Writer with header
    path = tmpdir_factory.mktemp("write_header").join("out.vcf.gz")
    header, _ = header_samples
    w = writer.Writer.from_path(path, header)
    # construct record to write out from scratch
    r = record.Record(
        "20",
        100,
        [],
        "C",
        [record.Substitution(record.SNV, "T")],
        None,
        [],
        O(),
        ["GT"],
        [
            record.Call("NA00001", O(GT="0/1")),
            record.Call("NA00002", O(GT="0/0")),
            record.Call("NA00003", O(GT="1/1")),
        ],
    )
    # write out the record, close file to ensure flushing to disk
    w.write_record(r)
    w.close()
    # check the resulting record
    LINE = "20\t100\t.\tC\tT\t.\t.\t.\tGT\t0/1\t0/0\t1/1\n"
    check_file(path, LINE)


def test_write_minimal_record_writer_from_stream_path(header_samples, tmpdir_factory):
    O = vcfpy.OrderedDict
    # open temporary file and setup the Writer with header
    path = tmpdir_factory.mktemp("write_header").join("out.vcf.gz")
    header, _ = header_samples
    with open(str(path), "wb") as f:
        w = writer.Writer.from_stream(f, header, path=str(path))
        # construct record to write out from scratch
        r = record.Record(
            "20",
            100,
            [],
            "C",
            [record.Substitution(record.SNV, "T")],
            None,
            [],
            O(),
            ["GT"],
            [
                record.Call("NA00001", O(GT="0/1")),
                record.Call("NA00002", O(GT="0/0")),
                record.Call("NA00003", O(GT="1/1")),
            ],
        )
        # write out the record, close file to ensure flushing to disk
        w.write_record(r)
        w.close()
    # check the resulting record
    LINE = "20\t100\t.\tC\tT\t.\t.\t.\tGT\t0/1\t0/0\t1/1\n"
    check_file(path, LINE)


def test_write_minimal_record_writer_from_stream_use_bgzf(header_samples, tmpdir_factory):
    O = vcfpy.OrderedDict
    # open temporary file and setup the Writer with header
    path = tmpdir_factory.mktemp("write_header").join("out.vcf.gz")
    header, samples = header_samples
    with open(str(path), "wb") as f:
        w = writer.Writer.from_stream(f, header, samples, use_bgzf=True)
        # construct record to write out from scratch
        r = record.Record(
            "20",
            100,
            [],
            "C",
            [record.Substitution(record.SNV, "T")],
            None,
            [],
            O(),
            ["GT"],
            [
                record.Call("NA00001", O(GT="0/1")),
                record.Call("NA00002", O(GT="0/0")),
                record.Call("NA00003", O(GT="1/1")),
            ],
        )
        # write out the record, close file to ensure flushing to disk
        w.write_record(r)
        w.close()
    # check the resulting record
    LINE = "20\t100\t.\tC\tT\t.\t.\t.\tGT\t0/1\t0/0\t1/1\n"
    check_file(path, LINE)



# -*- coding: utf-8 -*-
"""Test parsing of full VCF record lines
"""

import io
import sys

from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"


MEDIUM_HEADER = """
##fileformat=VCFv4.3
##fileDate=20090805
##source=myImputationProgramV3.1
##reference=file:///seq/references/1000GenomesPilot-NCBI36.fasta
##contig=<ID=20,length=62435964,assembly=B36,md5=f126cdf8a6e0c7f379d618ff66beb2da,species="Homo sapiens",taxonomy=x>
##phasing=partial
##INFO=<ID=NS,Number=1,Type=Integer,Description="Number of Samples With Data">
##INFO=<ID=DP,Number=1,Type=Integer,Description="Total Depth">
##INFO=<ID=AF,Number=A,Type=Float,Description="Allele Frequency">
##INFO=<ID=AA,Number=1,Type=String,Description="Ancestral Allele">
##INFO=<ID=DB,Number=0,Type=Flag,Description="dbSNP membership, build 129">
##INFO=<ID=H2,Number=0,Type=Flag,Description="HapMap2 membership">
##INFO=<ID=ANNO,Number=.,Type=String,Description="Additional annotation">
##FILTER=<ID=q10,Description="Quality below 10">
##FILTER=<ID=s50,Description="Less than 50% of samples have data">
##FILTER=<ID=FOO,Description="A test filter">
##FILTER=<ID=BAZ,Description="A second test filter">
##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">
##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality">
##FORMAT=<ID=DP,Number=1,Type=Integer,Description="Read Depth">
##FORMAT=<ID=HQ,Number=2,Type=Integer,Description="Haplotype Quality">
##FORMAT=<ID=FT,Number=1,Type=String,Description="Call-wise filters">
#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\tNA00001\tNA00002\tNA00003
""".lstrip()


def vcf_parser(lines):
    return parser.Parser(io.StringIO(MEDIUM_HEADER + lines), "<builtin>")


def test_parse_minimal_record():
    # Setup parser with stock header and lines to parse
    LINES = "20\t1\t.\tC\tG\t.\t.\t.\tGT\t0/1\t0/2\t.\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, [], OrderedDict(), ['GT'], "
            "[Call('NA00001', OrderedDict([('GT', '0/1')])), Call('NA00002', OrderedDict([('GT', '0/2')])), "
            "Call('NA00003', OrderedDict([('GT', None)]))])"
        )
    else:
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, [], {}, ['GT'], "
            "[Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/2'}), Call('NA00003', {'GT': None})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED


def test_parse_record_with_info():
    # Setup parser with stock header and lines to parse
    LINES = "20\t1\t.\tC\tG\t.\t.\tAA=G\tGT\t0/1\t0/1\t.\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, [], OrderedDict([('AA', 'G')]), "
            "['GT'], [Call('NA00001', OrderedDict([('GT', '0/1')])), Call('NA00002', OrderedDict([('GT', '0/1')])),"
            " Call('NA00003', OrderedDict([('GT', None)]))])"
        )
    else:
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, [], {'AA': 'G'}, "
            "['GT'], [Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/1'}), Call('NA00003', {'GT': None})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED


def test_parse_record_with_escaping():
    # Setup parser with stock header and lines to parse
    LINES = (
        "20\t100\t.\tC\tG\t.\t.\tANNO=Here%2Care%25some chars,"
        "%2525\tGT:FT\t0/1:FOO\t0/0:.\t1/1:.\n"
    )
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('20', 100, [], 'C', [Substitution(type_='SNV', value='G')], None, [], "
            "OrderedDict([('ANNO', ['Here,are%some chars', '%25'])]), ['GT', 'FT'], "
            "[Call('NA00001', OrderedDict([('GT', '0/1'), ('FT', ['FOO'])])),"
            " Call('NA00002', OrderedDict([('GT', '0/0'), ('FT', [])])),"
            " Call('NA00003', OrderedDict([('GT', '1/1'), ('FT', [])]))])"
        )
    else:
        EXPECTED = (
            "Record('20', 100, [], 'C', [Substitution(type_='SNV', value='G')], None, [], "
            "{'ANNO': ['Here,are%some chars', '%25']}, ['GT', 'FT'], "
            "[Call('NA00001', {'GT': '0/1', 'FT': ['FOO']}), Call('NA00002', {'GT': '0/0', 'FT': []}), "
            "Call('NA00003', {'GT': '1/1', 'FT': []})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED


def test_parse_record_with_filter_warning():
    # Setup parser with stock header and lines to parse
    LINES = "20\t1\t.\tC\tG\t.\tREX\t.\tGT:FT\t0/1:.\t0/2:BAR\t.:.\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, ['REX'], OrderedDict(), "
            "['GT', 'FT'], [Call('NA00001', OrderedDict([('GT', '0/1'), ('FT', [])])),"
            " Call('NA00002', OrderedDict([('GT', '0/2'), ('FT', ['BAR'])])),"
            " Call('NA00003', OrderedDict([('GT', None), ('FT', [])]))])"
        )
    else:
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, ['REX'], {}, "
            "['GT', 'FT'], [Call('NA00001', {'GT': '0/1', 'FT': []}),"
            " Call('NA00002', {'GT': '0/2', 'FT': ['BAR']}), Call('NA00003', {'GT': None, 'FT': []})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED


def test_parse_record_with_filter_no_warning():
    # Setup parser with stock header and lines to parse
    LINES = "20\t1\t.\tC\tG\t.\tREX\t.\tGT:FT\t0/1:.\t0/2:FOO;BAZ\t.:.\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, ['REX'], OrderedDict(), "
            "['GT', 'FT'], [Call('NA00001', OrderedDict([('GT', '0/1'), ('FT', [])])),"
            " Call('NA00002', OrderedDict([('GT', '0/2'), ('FT', ['FOO', 'BAZ'])])),"
            " Call('NA00003', OrderedDict([('GT', None), ('FT', [])]))])"
        )
    else:
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, ['REX'], {}, "
            "['GT', 'FT'], [Call('NA00001', {'GT': '0/1', 'FT': []}),"
            " Call('NA00002', {'GT': '0/2', 'FT': ['FOO', 'BAZ']}), Call('NA00003', {'GT': None, 'FT': []})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED


def test_missing_pass(recwarn):
    """Test parsing VCF file with missing ``PASS`` definition in header."""
    # Setup parser with stock header and lines to parse
    LINES = "20\t1\t.\tC\tG\t.\tPASS\t.\tGT\t0/1\t0/2\t.\n"
    p = vcf_parser(LINES)
    p.parse_header()
    # Perform the actual test
    if sys.version_info < (3, 6):
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, ['PASS'], OrderedDict(), ['GT'], "
            "[Call('NA00001', OrderedDict([('GT', '0/1')])), Call('NA00002', OrderedDict([('GT', '0/2')])), "
            "Call('NA00003', OrderedDict([('GT', None)]))])"
        )
    else:
        EXPECTED = (
            "Record('20', 1, [], 'C', [Substitution(type_='SNV', value='G')], None, ['PASS'], {}, ['GT'], "
            "[Call('NA00001', {'GT': '0/1'}), Call('NA00002', {'GT': '0/2'}), Call('NA00003', {'GT': None})])"
        )
    RESULT = p.parse_next_record()
    assert str(RESULT) == EXPECTED
    assert list(recwarn) == []



# -*- coding: utf-8 -*-
"""Parsing and converting of field values
"""

import pytest

from vcfpy import header
from vcfpy import parser

__author__ = "Manuel Holtgrewe <manuel.holtgrewe@bihealth.de>"

# parser.convert_field_value() ------------------------------------------------


def test_convert_field_values_integer_value():
    EXPECTED = 42
    RESULT = parser.convert_field_value("Integer", "42")
    assert EXPECTED == RESULT


def test_convert_field_values_integer_none():
    EXPECTED = None
    RESULT = parser.convert_field_value("Integer", ".")
    assert EXPECTED == RESULT


def test_convert_field_values_float_value():
    EXPECTED = 42.0
    RESULT = parser.convert_field_value("Float", "42")
    assert EXPECTED == RESULT


def test_convert_field_values_float_none():
    EXPECTED = None
    RESULT = parser.convert_field_value("Float", ".")
    assert EXPECTED == RESULT


def test_convert_field_values_flag():
    EXPECTED = True
    RESULT = parser.convert_field_value("Flag", True)
    assert EXPECTED == RESULT


def test_convert_field_values_character_value():
    EXPECTED = "X"
    RESULT = parser.convert_field_value("Character", "X")
    assert EXPECTED == RESULT


def test_convert_field_values_character_none():
    EXPECTED = None
    RESULT = parser.convert_field_value("Character", ".")
    assert EXPECTED == RESULT


def test_convert_field_values_string_value():
    EXPECTED = "Value"
    RESULT = parser.convert_field_value("String", "Value")
    assert EXPECTED == RESULT


def test_convert_field_values_string_none():
    EXPECTED = None
    RESULT = parser.convert_field_value("String", ".")
    assert EXPECTED == RESULT


# parser.parse_field_value() --------------------------------------------------


def test_parse_field_value_integer_one():
    EXPECTED = 42
    RESULT = parser.parse_field_value(header.FieldInfo("Integer", 1), "42")
    assert EXPECTED == RESULT


@pytest.mark.parametrize("number", [2, "A", "R", "G", "."])
def test_parse_field_value_integer_more(number):
    EXPECTED = [42, 43]
    RESULT = parser.parse_field_value(header.FieldInfo("Integer", number), "42,43")
    assert EXPECTED == RESULT


@pytest.mark.parametrize("number", [2, "A", "R", "G", "."])
def test_parse_field_value_integer_more_empty(number):
    EXPECTED = []
    RESULT = parser.parse_field_value(header.FieldInfo("Integer", number), ".")
    assert EXPECTED == RESULT


def test_parse_field_value_string_one():
    EXPECTED = "42"
    RESULT = parser.parse_field_value(header.FieldInfo("String", 1), "42")
    assert EXPECTED == RESULT


@pytest.mark.parametrize("number", [2, "A", "R", "G", "."])
def test_parse_field_value_string_more(number):
    EXPECTED = ["42", "43"]
    RESULT = parser.parse_field_value(header.FieldInfo("String", number), "42,43")
    assert EXPECTED == RESULT


@pytest.mark.parametrize("number", [2, "A", "R", "G", "."])
def test_parse_field_value_string_more_empty(number):
    EXPECTED = []
    RESULT = parser.parse_field_value(header.FieldInfo("String", number), ".")
    assert EXPECTED == RESULT
